#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
"""
The pyutilib.component.app package defines application interfaces that
simplify the use of the PyUtilib Component Architecture.
"""

from .simple import SimpleApplication



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

__all__ = ('Task', 'TaskProcessingError')

from pyutilib.pyro.util import using_pyro4

#
# Task returns a native Python data type.  This provides more
# more flexibility in the choice of serialization schemes than
# a user-defined class.  Also, the default serializer used by Pyro4
# ('serpent') does not support most user defined classes. The
# serializer can be set to 'pickle' by the user if they need
# this functionality (see: Pyro4 docs).
#


def Task(id=None, data=None, generateResponse=True):
    return {'id': id,
            'data': data,
            'result': None,
            'generateResponse': generateResponse,
            'processedBy': None,
            'client': None,
            'type': None}


#
# A simple yet identifiable type that indicates
# when an exception is encountered during task
# processing by a worker. We provide custome
# serialization hooks
#
class TaskProcessingError(Exception):

    def __init__(self, message):
        super(TaskProcessingError, self).__init__(message)

#
# register the special serialization hooks because
# the default Pyro4 serializer (Serpent) can not deal
# with custom classes
#
#
# Note: We can use this same mechanism to allow Task to be
#       re-implemented as a custom class.  HOWEVER,
#       TaskProcessingError is designed to be returned on the result
#       "slot" of Task, and when I've attempted to register
#       (de)serialization hooks for both Task and TaskProcessingError,
#       the end result is that the TaskProcessingError
#       de-serialization hook never gets called (leaving it in its
#       serialized dict form on the deserialized Task object). This
#       defeats the purpose of creating a TaskProcessingError type.
#       Perhaps there is a way around this, but for now I am giving
#       priority to having an error type that can be transmitted
#       for all serializers in both Pyro and Pyro4 (over making
#       Task a class again).
#
if using_pyro4:
    import Pyro4
    from Pyro4.util import SerializerBase

    # register hooks for TaskProcessingError
    def TaskProcessingError_to_dict(obj):
        return {"__class__": "pyutilib.pyro.task.TaskProcessingError",
                "message": obj.args[0]}

    def dict_to_TaskProcessingError(classname, d):
        return TaskProcessingError(d['message'])

    SerializerBase.register_class_to_dict(TaskProcessingError,
                                          TaskProcessingError_to_dict)
    SerializerBase.register_dict_to_class(
        "pyutilib.pyro.task.TaskProcessingError", dict_to_TaskProcessingError)



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________


class StreamIndenter(object):
    """Mock-up of a file-like object that wraps another file-like object
    and indents all data using the specified string before passing it to
    the underlying file.  Since this presents a full file interface,
    StreamIndenter objects may be arbitrarily nested."""

    def __init__(self, ostream, indent="        "):
        self.os = ostream
        self.indent = indent
        self.stripped_indent = indent.rstrip()
        self.newline = True

    def __getattr__(self, name):
        return getattr(self.os, name)

    def write(self, data):
        if not len(data):
            return
        lines = data.split('\n')
        if self.newline:
            if lines[0]:
                self.os.write(self.indent+lines[0])
            else:
                self.os.write(self.stripped_indent)
        else:
            self.os.write(lines[0])
        if len(lines) < 2:
            self.newline = False
            return
        for line in lines[1:-1]:
            if line:
                self.os.write("\n"+self.indent+line)
            else:
                self.os.write("\n"+self.stripped_indent)
        if lines[-1]:
            self.os.write("\n"+self.indent+lines[-1])
            self.newline = False
        else:
            self.os.write("\n")
            self.newline = True

    def writelines(self, sequence):
        for x in sequence:
            self.write(x)



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
"""
The outline of the PyUtilib Component Architecture (PCA) is adapted from Trac
(see the trac.core module).  This framework generalizes the Trac by supporting
multi-environment management of components, as well as non-singleton plugins.

This package provides a stand-alone module that defines all of the core
aspects of the PCA.  Related Python packages define extensions of this
framework that support current component-based applications.

NOTE: The PCA does not rely on any other part of PyUtilib.  Consequently,
this package can be independently used in other projects.
"""

from pyutilib.component.core.core import (
    Plugin, implements, Interface, CreatePluginFactory,
    PluginMeta, alias, ExtensionPoint, SingletonPlugin,
    PluginFactory, PluginError, PluginGlobals, with_metaclass,
    IPluginLoader, IPluginLoadPath, IIgnorePluginWhenLoading,
    IOptionDataProvider, PluginEnvironment
)

#
# This declaration is here because this is a convenient place where
# all symbols in this module have been defined.
#

class IgnorePluginPlugins(SingletonPlugin):
    """Ignore plugins from the pyutilib.component module"""

    implements(IIgnorePluginWhenLoading)

    def ignore(self, name):
        return name in globals()



import pyutilib.component.app
import os
import sys

currdir = sys.argv[-1] + os.sep

app = pyutilib.component.app.SimpleApplication("foo")
app.config.load(currdir + "log1.ini")
app.logger.log_dir = currdir
app.logger.log_file = "app3.log"
app.logger.reset_after_updates()
app.env.load_services()
app.log("test_app3 message")
app.flush()



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
#

import sys
import time
import traceback

__all__ = ('TicTocTimer', 'tic', 'toc')

#
# Setup the timer
#
if sys.version_info >= (3,3):
    # perf_counter is guaranteed to be monotonic and the most accurate timer
    _time_source = time.perf_counter
else:
    # On old Pythons, clock() is more accurate than time() on Windows
    # (.35us vs 15ms), but time() is more accurate than clock() on Linux
    # (1ns vs 1us).  It is unfortunate that time() is not monotonic, but
    # since the TicTocTimer is used for (potentially very accurate)
    # timers, we will sacrifice monotonicity on Linux for resolution.
    if sys.platform.startswith('win'):
        _time_source = time.clock
    else:
        _time_source = time.time

class TicTocTimer(object):
    """A class to calculate and report elapsed time.

    Examples:
       >>> from pyutilib.misc.timing import TicTocTimer
       >>> timer = TicTocTimer()
       >>> timer.tic('starting timer') # starts the elapsed time timer (from 0)
       >>> # ... do task 1
       >>> timer.toc('task 1') # prints the elapsed time for task 1

    If no ostream or logger is provided, then output is printed to sys.stdout

    Args:
        ostream (FILE): an optional output stream to print the timing
            information
        logger (Logger): an optional output stream using the python
           logging package. Note: timing logged using logger.info
    """
    def __init__(self, ostream=None, logger=None):
        self._lastTime = self._loadTime = _time_source()
        self._ostream = ostream
        self._logger = logger
        self._start_count = 0
        self._cumul = 0

    def tic(self, msg=None, ostream=None, logger=None):
        """Reset the tic/toc delta timer.

        This resets the reference time from which the next delta time is
        calculated to the current time.

        Args:
            msg (str): The message to print out.  If :const:`None`
                (default), then prints out "Resetting the tic/toc delta
                timer"; if it evaluates to :const:`False` (:const:`0`,
                :const:`False`, :const:`""`) then no message is printed.
            ostream (FILE): an optional output stream (overrides the ostream
                provided when the class was constructed).
            logger (Logger): an optional output stream using the python
                logging package (overrides the ostream provided when the
                class was constructed). Note: timing logged using logger.info
        """
        self._lastTime = _time_source()
        if msg is None:
            msg = "Resetting the tic/toc delta timer"
        if msg:
            self.toc(msg=msg, delta=False, ostream=ostream, logger=logger)


    def toc(self, msg=None, delta=True, ostream=None, logger=None):
        """Print out the elapsed time.

        This resets the reference time from which the next delta time is
        calculated to the current time.

        Args:
            msg (str): The message to print out.  If :const:`None`
                (default), then print out the file name, line number,
                and function that called this method; if it evaluates to
                :const:`False` (:const:`0`, :const:`False`, :const:`""`)
                then no message is printed.
            delta (bool): print out the elapsed wall clock time since
                the last call to :meth:`tic` or :meth:`toc`
                (:const:`True` (default)) or since the module was first
                loaded (:const:`False`).
            ostream (FILE): an optional output stream (overrides the ostream
                provided when the class was constructed).
            logger (Logger): an optional output stream using the python
                logging package (overrides the ostream provided when the
                class was constructed). Note: timing logged using logger.info
        """

        if msg is None:
            msg = 'File "%s", line %s in %s' % \
                  traceback.extract_stack(limit=2)[0][:3]

        now = _time_source()
        if self._start_count or self._lastTime is None:
            ans = self._cumul
            if self._lastTime:
                ans += _time_source() - self._lastTime
            if msg:
                msg = "[%8.2f|%4d] %s\n" % (ans, self._start_count, msg)
        elif delta:
            ans = now - self._lastTime
            self._lastTime = now
            if msg:
                msg = "[+%7.2f] %s\n" % (ans, msg)
        else:
            ans = now - self._loadTime
            if msg:
                msg = "[%8.2f] %s\n" % (ans, msg)

        if msg:
            if ostream is None:
                ostream = self._ostream
                if ostream is None and logger is None:
                    ostream = sys.stdout
            if ostream is not None:
                ostream.write(msg)

            if logger is None:
                logger = self._logger
            if logger is not None:
                logger.info(msg)

        return ans

    def stop(self):
        try:
            delta = _time_source() - self._lastTime
        except TypeError:
            if self._lastTime is None:
                raise RuntimeError(
                    "Stopping a TicTocTimer that was already stopped")
            raise
        self._cumul += delta
        self._lastTime = None
        return delta

    def start(self):
        if self._lastTime:
            self.stop()
        self._start_count += 1
        self._lastTime = _time_source()

_globalTimer = TicTocTimer()
tic = _globalTimer.tic
toc = _globalTimer.toc


class _HierarchicalHelper(object):
    def __init__(self):
        self.tic_toc = TicTocTimer()
        self.timers = dict()
        self.total_time = 0
        self.n_calls = 0

    def start(self):
        self.n_calls += 1
        self.tic_toc.start()

    def stop(self):
        self.total_time += self.tic_toc.stop()

    def to_str(self, indent, stage_identifier_lengths):
        s = ''
        if len(self.timers) > 0:
            underline = indent + '-' * (sum(stage_identifier_lengths) + 36) + '\n'
            s += underline
            name_formatter = '{name:<' + str(sum(stage_identifier_lengths)) + '}'
            other_time = self.total_time
            sub_stage_identifier_lengths = stage_identifier_lengths[1:]
            for name, timer in self.timers.items():
                if self.total_time > 0:
                    _percent = timer.total_time / self.total_time * 100
                else:
                    _percent = float('nan')
                s += indent
                s += ( name_formatter + '{ncalls:>9d} {cumtime:>9.3f} '
                       '{percall:>9.3f} {percent:>6.1f}\n' ).format(
                           name=name,
                           ncalls=timer.n_calls,
                           cumtime=timer.total_time,
                           percall=timer.total_time/timer.n_calls,
                           percent=_percent )
                s += timer.to_str(
                    indent=indent + ' '*stage_identifier_lengths[0],
                    stage_identifier_lengths=sub_stage_identifier_lengths)
                other_time -= timer.total_time

            if self.total_time > 0:
                _percent = other_time / self.total_time * 100
            else:
                _percent = float('nan')
            s += indent
            s += ( name_formatter + '{ncalls:>9} {cumtime:>9.3f} '
                   '{percall:>9} {percent:>6.1f}\n' ).format(
                       name='other',
                       ncalls='n/a',
                       cumtime=other_time,
                       percall='n/a',
                       percent=_percent )
            s += underline.replace('-', '=')
        return s


class HierarchicalTimer(object):
    """A class for hierarchical timing.

    Examples
    --------
    >>> import time
    >>> from pyutilib.misc.timing import HierarchicalTimer
    >>> timer = HierarchicalTimer()
    >>> timer.start('all')
    >>> time.sleep(0.2)
    >>> for i in range(10):
    ...     timer.start('a')
    ...     time.sleep(0.1)
    ...     for i in range(5):
    ...         timer.start('aa')
    ...         time.sleep(0.01)
    ...         timer.stop('aa')
    ...     timer.start('ab')
    ...     timer.stop('ab')
    ...     timer.stop('a')
    ...
    >>> for i in range(10):
    ...     timer.start('b')
    ...     time.sleep(0.02)
    ...     timer.stop('b')
    ...
    >>> timer.stop('all')
    >>> print(timer)
    Identifier        ncalls   cumtime   percall      %
    ---------------------------------------------------
    all                    1     2.248     2.248  100.0
         ----------------------------------------------
         a                10     1.787     0.179   79.5
              -----------------------------------------
              aa          50     0.733     0.015   41.0
              ab          10     0.000     0.000    0.0
              other      n/a     1.055       n/a   59.0
              =========================================
         b                10     0.248     0.025   11.0
         other           n/a     0.213       n/a    9.5
         ==============================================
    ===================================================
    <BLANKLINE>

    The columns are:
      ncalls : The number of times the timer was started and stopped
      cumtime: The cumulative time (in seconds) the timer was active
               (started but not stopped)
      percall: cumtime (in seconds) / ncalls
      %      : This is cumtime of the timer divided by cumtime of the
               parent timer times 100


    >>> print('a total time: %f' % timer.get_total_time('all.a'))
    a total time: 1.902037
    >>> print('ab num calls: %d' % timer.get_num_calls('all.a.ab'))
    ab num calls: 10
    >>> print('aa %% time: %f' % timer.get_relative_percent_time('all.a.aa'))
    aa % time: 44.144148
    >>> print('aa %% total: %f' % timer.get_total_percent_time('all.a.aa'))
    aa % total: 35.976058

    Internal Workings
    -----------------
    The HierarchicalTimer use a stack to track which timers are active
    at any point in time. Additionally, each timer has a dictionary of
    timers for its children timers. Consider

    >>> timer = HierarchicalTimer()
    >>> timer.start('all')
    >>> timer.start('a')
    >>> timer.start('aa')

    After the above code is run, self.stack will be ['all', 'a', 'aa']
    and self.timers will have one key, 'all' and one value which will be
    a _HierarchicalHelper. The _HierarchicalHelper has its own timers
    dictionary:

    {'a': _HierarchicalHelper}

    and so on. This way, we can easily access any timer with something
    that looks like the stack. The logic is recursive (although the
    code is not).

    """
    def __init__(self):
        self.stack = list()
        self.timers = dict()

    def _get_timer(self, identifier, should_exist=False):
        """

        This method gets the timer associated with the current state
        of self.stack and the specified identifier.

        Parameters
        ----------
        identifier: str
            The name of the timer
        should_exist: bool
            The should_exist is True, and the timer does not already
            exist, an error will be raised. If should_exist is False, and
            the timer does not already exist, a new timer will be made.

        Returns
        -------
        timer: _HierarchicalHelper

        """
        parent = self._get_timer_from_stack(self.stack)
        if identifier in parent.timers:
            return parent.timers[identifier]
        else:
            if should_exist:
                raise RuntimeError(
                    'Could not find timer {0}'.format(
                        '.'.join(self.stack + [identifier])))
            parent.timers[identifier] = _HierarchicalHelper()
            return parent.timers[identifier]

    def start(self, identifier):
        """
        Start incrementing the timer identified with identifier

        Parameters
        ----------
        identifier: str
            The name of the timer
        """
        timer = self._get_timer(identifier)
        timer.start()
        self.stack.append(identifier)

    def stop(self, identifier):
        """
        Stop incrementing the timer identified with identifier

        Parameters
        ----------
        identifier: str
            The name of the timer
        """
        if identifier != self.stack[-1]:
            raise ValueError(
                str(identifier) + ' is not the currently active timer.  '
                'The only timer that can currently be stopped is '
                + '.'.join(self.stack))
        self.stack.pop()
        timer = self._get_timer(identifier, should_exist=True)
        timer.stop()

    def _get_identifier_len(self):
        stage_timers = list(self.timers.items())
        stage_lengths = list()

        while len(stage_timers) > 0:
            new_stage_timers = list()
            max_len = 0
            for identifier, timer in stage_timers:
                new_stage_timers.extend(timer.timers.items())
                if len(identifier) > max_len:
                    max_len = len(identifier)
            stage_lengths.append(max(max_len, len('other')))
            stage_timers = new_stage_timers

        return stage_lengths

    def __str__(self):
        stage_identifier_lengths = self._get_identifier_len()
        name_formatter = '{name:<' + str(sum(stage_identifier_lengths)) + '}'
        s = ( name_formatter + '{ncalls:>9} {cumtime:>9} '
              '{percall:>9} {percent:>6}\n').format(
                  name='Identifier',
                  ncalls='ncalls',
                  cumtime='cumtime',
                  percall='percall',
                  percent='%')
        underline = '-' * (sum(stage_identifier_lengths) + 36) + '\n'
        s += underline
        sub_stage_identifier_lengths = stage_identifier_lengths[1:]
        for name, timer in self.timers.items():
            s += ( name_formatter + '{ncalls:>9d} {cumtime:>9.3f} '
                   '{percall:>9.3f} {percent:>6.1f}\n').format(
                       name=name,
                       ncalls=timer.n_calls,
                       cumtime=timer.total_time,
                       percall=timer.total_time/timer.n_calls,
                       percent=self.get_total_percent_time(name))
            s += timer.to_str(
                indent=' '*stage_identifier_lengths[0],
                stage_identifier_lengths=sub_stage_identifier_lengths)
        s += underline.replace('-', '=')
        return s

    def reset(self):
        """
        Completely reset the timer.
        """
        self.stack = list()
        self.timers = dict()

    def _get_timer_from_stack(self, stack):
        """
        This method gets the timer associated with stack.

        Parameters
        ----------
        stack: list of str
            A list of identifiers.

        Returns
        -------
        timer: _HierarchicalHelper
        """
        tmp = self
        for i in stack:
            tmp = tmp.timers[i]
        return tmp

    def get_total_time(self, identifier):
        """
        Parameters
        ----------
        identifier: str
            The full name of the timer including parent timers separated
            with dots.

        Returns
        -------
        total_time: float
            The total time spent with the specified timer active.
        """
        stack = identifier.split('.')
        timer = self._get_timer_from_stack(stack)
        return timer.total_time

    def get_num_calls(self, identifier):
        """
        Parameters
        ----------
        identifier: str
            The full name of the timer including parent timers separated
            with dots.

        Returns
        -------
        num_calss: int
            The number of times start was called for the specified timer.
        """
        stack = identifier.split('.')
        timer = self._get_timer_from_stack(stack)
        return timer.n_calls

    def get_relative_percent_time(self, identifier):
        """
        Parameters
        ----------
        identifier: str
            The full name of the timer including parent timers separated
            with dots.

        Returns
        -------
        percent_time: float
            The percent of time spent in the specified timer
            relative to the timer's immediate parent.
        """
        stack = identifier.split('.')
        timer = self._get_timer_from_stack(stack)
        parent = self._get_timer_from_stack(stack[:-1])
        if parent is self:
            return self.get_total_percent_time(identifier)
        else:
            if parent.total_time > 0:
                return timer.total_time / parent.total_time * 100
            else:
                return float('nan')

    def get_total_percent_time(self, identifier):
        """
        Parameters
        ----------
        identifier: str
            The full name of the timer including parent timers separated
            with dots.

        Returns
        -------
        percent_time: float
            The percent of time spent in the specified timer
            relative to the total time in all timers.
        """
        stack = identifier.split('.')
        timer = self._get_timer_from_stack(stack)
        total_time = 0
        for _timer in self.timers.values():
            total_time += _timer.total_time
        if total_time > 0:
            return timer.total_time / total_time * 100
        else:
            return float('nan')



import sys
import time
# Write ERR first so that the output is deterministic
sys.stderr.write("Tee Script: ERR\n")
sys.stdout.write("Tee Script: OUT\n")
# sleep just enough so that the STDERR thread should finish flushing its
# buffer before the process terminates (at which point the order in
# which the buffers are flushed is somewhat random).
time.sleep(0.5)



import sys

__all__ = ['SparseMapping']

import six

try:
    if six.PY2:
        from collections import MutableMapping
    else:
        from collections.abc import MutableMapping

    class SparseMapping(MutableMapping):
        """
        Class that provides a default value for keys that are not defined in
        the dictionary, a specified index for keys, and a specified value that
        mappings are within.
        """

        def __init__(self, default=None, index=None, within=None, *args,
                     **kwds):
            self._map = kwds
            self.default = default
            self._index = index
            self.within = within
            self.update(dict(*args, **kwds))

        def nondefault_keys(self):
            return self._map.keys()

        def nondefault_iter(self):
            return iter(self._map)

        def __len__(self):
            if self._index is None or self.default is None:
                return len(self._map)
            return len(self._index)

        def __contains__(self, key):
            if self._index is None:
                return key in self._map
            return key in self._index

        def __iter__(self):
            if self._index is None or self.default is None:
                return iter(self._map)
            return iter(self._index)

        def __getitem__(self, key):
            if key in self._map:
                return self._map[key]
            if not self.default is None and (self._index is None or
                                             key in self._index):
                return self.default
            if (self.default is None) and (not self._index is None) and (
                    key in self._index):
                raise ValueError(
                    "Legal key '%s' specified in SparseMapping, but value is uninitialized and there is no default value"
                    % key)
            raise KeyError("Unknown key value: %s" % str(key))

        def __setitem__(self, key, value):
            if not self._index is None and not key in self._index:
                raise KeyError("Unknown key value: %s" % str(key))
            if not self.within is None and not value in self.within:
                raise ValueError("Bad mapping value: %s" % str(value))
            self._map[key] = value

        def index(self):
            return self._index

        def set_item(self, key, value):
            self._map[key] = value

        def __delitem__(self, key):
            del self._map[key]

except ImportError:

    class SparseMapping(dict):
        """
        Class that provides a default value for keys that are not defined in
        the dictionary.

        Adapted from code developed by Andreas Kloss and submitted to
        the ActiveState Programmer Network http://aspn.activestate.com
        """

        def __init__(self, default=None, index=None, within=None, *args,
                     **kwds):
            super(SparseMapping, self).__init__()
            self.update(kwds)
            self.default = default
            self._index = index
            self.within = within

        def nondefault_keys(self):
            return self.keys()

        def nondefault_iter(self):
            return iter(self)

        def set_item(self, key, value):
            self[key] = value

        def index(self):
            return self._index

        def __len__(self):
            if self._index is None or self.default is None:
                return dict.__len__(self)
            return len(self._index)

        def __getitem__(self, key):
            try:
                return dict.__getitem__(self, key)
            except KeyError:
                err = sys.exc_info()[1]
                if not self.default is None and (self._index is None or
                                                 key in self._index):
                    return self.default
                raise err

        def __copy__(self):
            return SparseMapping(self.default, self)



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

__all__ = ['TaskPlugin', 'TaskFactory', 'WorkflowPlugin']

from pyutilib.workflow import task
from pyutilib.workflow import workflow
from pyutilib.component.core import Plugin, implements, Interface, CreatePluginFactory, alias


class IWorkflowTask(Interface):
    pass


TaskFactory = CreatePluginFactory(IWorkflowTask)


class TaskPlugin(Plugin, task.Task):

    implements(IWorkflowTask)

    def __init__(self, *args, **kwds):  #pragma:nocover
        Plugin.__init__(self, *args, **kwds)
        task.Task.__init__(self, *args, **kwds)

    def __repr__(self):
        return task.Task.__repr__(self)  #pragma:nocover


class WorkflowPlugin(Plugin, workflow.Workflow):

    implements(IWorkflowTask)

    def __init__(self, *args, **kwds):  #pragma:nocover
        Plugin.__init__(self, *args, **kwds)
        workflow.Workflow.__init__(self, *args, **kwds)

    def __repr__(self):  #pragma:nocover
        return workflow.Workflow.__repr__(self)


class Selection_Task(TaskPlugin):

    alias('workflow.selection')

    def __init__(self, *args, **kwds):
        TaskPlugin.__init__(self, *args, **kwds)
        #
        self.inputs.declare('index')
        self.inputs.declare('data')
        #
        self.outputs.declare('selection')

    def execute(self):
        self.selection = self.data[self.index]


class Switch_Task(TaskPlugin):

    alias('workflow.switch')

    def __init__(self, *args, **kwds):
        TaskPlugin.__init__(self, *args, **kwds)
        self._branches = {}
        self.inputs.declare('value')

    def add_branch(self, value, task):
        self._branches[value] = 'task' + str(task.id)
        self.output_controls.declare(self._branches[value])

        task.input_controls.declare('task' + str(self.id))
        setattr(task.input_controls, 'task' + str(self.id),
                self.output_controls[self._branches[value]])

    def execute(self):
        flag = False
        for key in self._branches:
            if self.value == key:
                self.output_controls[self._branches[key]].set_ready()
                flag = True
            else:
                self.output_controls[self._branches[key]].reset()
        if not flag:
            raise ValueError(
                "Branch condition has value '%s' but no branch is indexed with that value.\n    Valid branch indices: %s"
                % (str(self.value), sorted(self._branches.keys())))

    def __repr__(self):
        return TaskPlugin.__repr__(self)  #pragma:nocover


class IfThen_Task(Switch_Task):

    alias('workflow.branch')

    def __init__(self, *args, **kwds):
        Switch_Task.__init__(self, *args, **kwds)

    def __repr__(self):
        return Switch_Task.__repr__(self)  #pragma:nocover



import os
import sys
from os.path import abspath, dirname
currdir = dirname(abspath(__file__)) + os.sep

import pyutilib.th as unittest
import pyutilib.workflow
import pyutilib.workflow.globals


class DummyResource(pyutilib.workflow.Resource):

    def __init__(self, name=None):
        pyutilib.workflow.Resource.__init__(self)

    def available(self):
        return False


class TaskA(pyutilib.workflow.Task):

    def __init__(self, *args, **kwds):
        pyutilib.workflow.Task.__init__(self, *args, **kwds)
        self.inputs.declare('x')
        self.inputs.declare('y')
        self.outputs.declare('z')

    def execute(self):
        self.z = self.x + self.y


class TaskAA1(pyutilib.workflow.Task):

    def __init__(self, *args, **kwds):
        pyutilib.workflow.Task.__init__(self, *args, **kwds)
        self.inputs.declare('x', action='append')
        self.outputs.declare('z')

    def execute(self):
        self.z = sum(self.x)


class TaskAA2(pyutilib.workflow.Task):

    def __init__(self, *args, **kwds):
        pyutilib.workflow.Task.__init__(self, *args, **kwds)
        self.inputs.declare('x', action='map')
        self.outputs.declare('z')

    def execute(self):
        self.z = sum(self.x.keys())


class TaskAA3(pyutilib.workflow.Task):

    def __init__(self, *args, **kwds):
        pyutilib.workflow.Task.__init__(self, *args, **kwds)
        self.inputs.declare('x', action='map_any')
        self.outputs.declare('z')

    def execute(self):
        self.z = sum(self.x.keys())


class TaskB(pyutilib.workflow.Task):

    def __init__(self, *args, **kwds):
        pyutilib.workflow.Task.__init__(self, *args, **kwds)
        self.inputs.declare('a')
        self.inputs.declare('i')
        self.outputs.declare('b')

    def execute(self):
        self.b = 100 * self.i + 2 * self.a


class TaskAAA(pyutilib.workflow.Task):

    def __init__(self, *args, **kwds):
        pyutilib.workflow.Task.__init__(self, *args, **kwds)
        self.inputs.declare('a', optional=True)
        self.outputs.declare('a', self.inputs.a)

    def execute(self):
        pass


class TaskBB(TaskB):

    def __init__(self, *args, **kwds):
        TaskB.__init__(self, *args, **kwds)

    def _create_parser(self, parser=None):
        pyutilib.workflow.Task._create_parser(self, parser)
        self.add_argument("--a", dest="a", type=int)
        self.add_argument("--i", dest="i", type=int)


class TaskC(pyutilib.workflow.Task):

    def __init__(self, *args, **kwds):
        pyutilib.workflow.Task.__init__(self, *args, **kwds)
        self.inputs.declare('i')
        self.outputs.declare('o')

    def execute(self):
        self.o = 10 * self.i


class TaskCC(TaskC):

    def __init__(self, *args, **kwds):
        TaskC.__init__(self, *args, **kwds)

    def _create_parser(self, parser=None):
        pyutilib.workflow.Task._create_parser(self, parser)
        self.add_argument("--i", dest="i", type=int)


class TaskD(pyutilib.workflow.Task):

    def __init__(self, *args, **kwds):
        pyutilib.workflow.Task.__init__(self, *args, **kwds)
        self.inputs.declare('i')
        self.outputs.declare('o')
        self.outputs.declare('z')

    def execute(self):
        self.o = 10 * self.i
        self.z = -999


class TaskE(pyutilib.workflow.Task):

    def __init__(self, *args, **kwds):
        pyutilib.workflow.Task.__init__(self, *args, **kwds)
        self.inputs.declare('I')
        self.outputs.declare('O')

    def execute(self):
        self.O = self.I


class Test(unittest.TestCase):

    def test1(self):
        pyutilib.workflow.globals.reset_id_counter()
        # Define tasks
        A = TaskA()
        B = TaskB()
        C = TaskC()
        # Establish connections
        A.inputs.x = B.outputs.b
        A.inputs.y = C.outputs.o
        # Create workflow
        w = pyutilib.workflow.Workflow()
        w.add(A)
        w.add(
            A)  # This is redundant, but it double-checks the logic for adding tasks
        #
        # Print workflow
        #
        ##print "..."
        self.assertEqual(str(w),\
"""Workflow Task4:
Task5 prev: [] next: [2, 3] resources: []
Task2 prev: [5] next: [1] resources: []
Task3 prev: [5] next: [1] resources: []
Task1 prev: [2, 3] next: [6] resources: []
Task6 prev: [1] next: [] resources: []""")
        self.assertEqual(w(i=3, a=2), {'z': 334})

    def test1c(self):
        pyutilib.workflow.globals.reset_id_counter()
        # Define tasks
        A = TaskA()
        E1 = TaskE()
        E2 = TaskE()
        C = TaskC()
        # Establish connections
        A.inputs.x = E1.outputs.O
        A.inputs.y = E2.outputs.O
        E1.inputs.I = C.outputs.o
        E2.inputs.I = C.outputs.o
        # Create workflow
        w = pyutilib.workflow.Workflow()
        w.add(A)
        #
        # Print workflow
        #
        ##print "..."
        self.assertEqual(str(w),\
"""Workflow Task5:
Task6 prev: [] next: [4] resources: []
Task4 prev: [6] next: [2, 3] resources: []
Task2 prev: [4] next: [1] resources: []
Task3 prev: [4] next: [1] resources: []
Task1 prev: [2, 3] next: [7] resources: []
Task7 prev: [1] next: [] resources: []""")
        self.assertEqual(w(i=3), {'z': 60})

    def test1b(self):
        pyutilib.workflow.globals.reset_id_counter()
        # Define tasks
        A = TaskA()
        B = TaskB()
        C = TaskC()
        # Establish connections
        A.inputs.x = B.outputs.b
        A.inputs.y = C.outputs.o
        # Create workflow
        w = pyutilib.workflow.Workflow()
        w.add(A, loadall=False)
        self.assertEqual(
            str(w), "Workflow Task4:\nTask5 prev: [] next: [] resources: []")

    def test1a(self):
        pyutilib.workflow.globals.reset_id_counter()
        # Create workflow
        w = pyutilib.workflow.Workflow()
        self.assertEqual(
            str(w), "Workflow Task1:\nTask2 prev: [] next: [] resources: []")
        w.add(pyutilib.workflow.task.NoTask)
        self.assertEqual(
            str(w), "Workflow Task1:\nTask2 prev: [] next: [] resources: []")

    @unittest.skipIf(sys.version_info >= (3, 0), "There is a slight (space) formatting differences in different Python version... Skipping test.") # yapf: disable
    def test2(self):
        # Do we really want to be testing pformat output?  I think we might
        # actually want to override __cmp__ in the workflow.Task code and instead
        # use that.
        pyutilib.workflow.globals.reset_id_counter()
        A = TaskA(name="A")
        B = TaskB()
        A.inputs.x = B.outputs.b
        base = """{ 'A_TYPE': 'Task',
  'Id': 1,
  'InputControls': { 'A_TYPE': 'Port',
                     'Mode': 'inputs',
                     'Name': 'A-input-controls',
                     'Owner': 'A prev: [2] next: [] resources: []'},
  'Inputs': { 'A_TYPE': 'Port',
              'Mode': 'inputs',
              'Name': 'A-inputs',
              'Owner': 'A prev: [2] next: [] resources: []',
              'x': { 'A_TYPE': 'Port',
                     'Action': 'store',
                     'Connections': { 'Inputs': [ 'DirectConnector: from=(2) to=(1) False'],
                                      'Outputs': []},
                     'Constant': 'False',
                     'Name': 'x',
                     'Optional': 'False',
                     'Ready': 'False',
                     'Task': '1',
                     'Value': 'None'},
              'y': { 'A_TYPE': 'Port',
                     'Action': 'store',
                     'Connections': { 'Inputs': [], 'Outputs': []},
                     'Constant': 'False',
                     'Name': 'y',
                     'Optional': 'False',
                     'Ready': 'False',
                     'Task': '1',
                     'Value': 'None'}},
  'Name': 'A',
  'OutputControls': { 'A_TYPE': 'Port',
                      'Mode': 'outputs',
                      'Name': 'A-output-controls',
                      'Owner': 'A prev: [2] next: [] resources: []'},
  'Outputs': { 'A_TYPE': 'Port',
               'Mode': 'outputs',
               'Name': 'A-outputs',
               'Owner': 'A prev: [2] next: [] resources: []',
               'z': { 'A_TYPE': 'Port',
                      'Action': 'store',
                      'Connections': { 'Inputs': [], 'Outputs': []},
                      'Constant': 'False',
                      'Name': 'z',
                      'Optional': 'False',
                      'Ready': 'False',
                      'Task': '1',
                      'Value': 'None'}}}"""
        self.assertEqual(str(A), base)

    def test3(self):
        # Do we really want to be testing pformat output?  I think we might
        # actually want to override __cmp__ in the workflow.Task code and instead
        # use that.
        pyutilib.workflow.globals.reset_id_counter()
        A = TaskA()
        B = TaskB()
        A.inputs.x = B.outputs.b
        OUTPUT = open(currdir + 'test3.out', 'w')
        OUTPUT.write(str(A.inputs) + '\n')
        OUTPUT.write(str(A.outputs) + '\n')
        OUTPUT.close()
        self.assertFileEqualsBaseline(currdir + 'test3.out',
                                      currdir + 'test3.txt')

    def test4(self):
        # Do we really want to be testing pformat output?  I think we might
        # actually want to override __cmp__ in the workflow.Task code and instead
        # use that.
        pyutilib.workflow.globals.reset_id_counter()
        A = TaskA()
        A.inputs['x'] = 2
        self.assertEqual(A.inputs['x'].get_value(), 2)
        OUTPUT = open(currdir + 'test4.out', 'w')
        OUTPUT.write(str(A) + '\n')
        OUTPUT.close()
        self.assertFileEqualsBaseline(currdir + 'test4.out',
                                      currdir + 'test4.txt')

    def test5(self):
        pyutilib.workflow.globals.reset_id_counter()
        # Define tasks
        A = TaskA()
        B = TaskB()
        C = TaskC()
        # Establish connections
        A.inputs.x = B.outputs.b
        A.inputs.y = C.outputs.o
        # Create workflow
        w = pyutilib.workflow.Workflow()
        w.add(A)
        w.set_options(['--i=3', '--a=2'])
        try:
            w()
            self.fail("Expected ValueError because the inputs are not defined")
        except ValueError:
            pass

    def test5a(self):
        pyutilib.workflow.globals.reset_id_counter()
        # Define tasks
        A = TaskA()
        B = TaskBB()
        C = TaskCC()
        # Establish connections
        A.inputs.x = B.outputs.b
        A.inputs.y = C.outputs.o
        # Create workflow
        w = pyutilib.workflow.Workflow()
        w.add(A)
        w.set_options(['--i=3', '--a=2'])
        self.assertEqual(w(), {'z': 334})
        # This doesn't work, since we can't merge options that aren't
        # in groups.  The soln is to move to argparse, but I'll save that for later
        #w.print_help()

    def test5b(self):
        # Do we really want to be testing pformat output?  I think we might
        # actually want to override __cmp__ in the workflow.Task code and instead
        # use that.
        self.skipTest("This test is not portable to different Python version")

        pyutilib.workflow.globals.reset_id_counter()
        # Define tasks
        A = TaskA()
        B = TaskBB()
        C = TaskCC()
        # Establish connections
        A.inputs.x = B.outputs.b
        A.inputs.y = C.outputs.o
        # Create workflow
        w = pyutilib.workflow.Workflow()
        w.add(A)
        w.set_options(['--i=3', '--a=2'])
        w()
        OUTPUT = open(currdir + 'test5b.out', 'w')
        OUTPUT.write(str(w) + '\n')
        OUTPUT.close()
        self.assertFileEqualsBaseline(currdir + 'test5b.out',
                                      currdir + 'test5b.txt')

    def test_error1(self):
        pyutilib.workflow.globals.reset_id_counter()
        A = TaskA()
        try:
            A.inputs['inputs'] = 2
            self.fail(
                "Expected ValueError because we're using the reserved 'inputs' attribute name")
        except TypeError:
            pass

    def test_error2(self):
        pyutilib.workflow.globals.reset_id_counter()
        A = TaskA()
        try:
            print(A.inputs._foo)
            self.fail("Should have generated error for unknown attribute _foo")
        except AttributeError:
            pass

    def test_error3(self):
        pyutilib.workflow.globals.reset_id_counter()
        # Define tasks
        A = TaskA()
        B = TaskB()
        D = TaskD()
        # Establish connections
        A.inputs.x = B.outputs.b
        A.inputs.y = D.outputs.o
        # Create workflow
        w = pyutilib.workflow.Workflow()
        try:
            w.add(A)
            self.fail(
                "Expected ValueError because there are multiple outputs with the same name")
        except ValueError:
            pass

    def test_error4(self):
        pyutilib.workflow.globals.reset_id_counter()
        # Define tasks
        A = TaskAA1()
        B = TaskAAA()
        C = TaskAAA()
        # Establish connections
        A.inputs.x = B.outputs.a
        A.inputs.x = C.outputs.a
        # Create workflow
        w = pyutilib.workflow.Workflow()
        w.add(A)
        try:
            w()
            self.fail("Expected ValueError because inputs are None")
        except ValueError:
            pass

    def test_error5(self):
        pyutilib.workflow.globals.reset_id_counter()
        # Define tasks
        A = TaskAA2()
        B = TaskAAA()
        C = TaskAAA()
        # Establish connections
        A.inputs.x = B.outputs.a
        A.inputs.x = C.outputs.a
        # Create workflow
        w = pyutilib.workflow.Workflow()
        w.add(A)
        try:
            w()
            self.fail("Expected ValueError because inputs are None")
        except ValueError:
            pass

    def test_error6(self):
        driver = pyutilib.workflow.TaskDriver()
        try:
            driver.register_task('_foo_')
            self.fail("Expected ValueError because '_foo_' task is not defined")
        except ValueError:
            pass

    def test_error7(self):
        A1 = TaskA()
        A2 = TaskA()
        A3 = TaskA()
        A3.inputs.x = A1.outputs.z
        try:
            A3.inputs.x = A2.outputs.z
            self.fail(
                "Expected ValueError because we're connecting two outputs to an input with action 'store'")
        except:
            pass

    def test_error8(self):
        A1 = TaskAA1()
        A2 = TaskAA1()
        A2.inputs.x = A1.outputs.z
        try:
            A2.inputs.x.validate()
            self.fail(
                "Expected ValueError because the 'x' input does not have a nontrivial value")
        except:
            pass
        self.assertFalse(A2.ready())

    def test_error9(self):
        A1 = TaskAA2()
        A2 = TaskAA2()
        A2.inputs.x = A1.outputs.z
        try:
            A2.inputs.x.validate()
            self.fail(
                "Expected ValueError because the 'x' input does not have a nontrivial value")
        except:
            pass
        self.assertFalse(A2.ready())

    def test_error10(self):
        A1 = TaskAA3()
        A2 = TaskAA3()
        A3 = TaskAA3()
        A2.inputs.x = A1.outputs.z
        A2.inputs.x = A3.outputs.z
        try:
            A2.inputs.x.validate()
            self.fail(
                "Expected ValueError because the 'x' input does not have a nontrivial value")
        except:
            pass
        self.assertFalse(A2.ready())
        A1.outputs.z = 1
        self.assertTrue(A2.ready())

    def test_error11(self):
        A1 = TaskA()
        A1.add_resource(DummyResource())
        try:
            A1()
            self.fail("Expected IOError because resource is busy.")
        except IOError:
            pass

    def test_driver1(self):
        driver = pyutilib.workflow.TaskDriver()
        driver.register_task('workflow.selection')
        try:
            driver.parse_args(['-h'])
            self.fail("Expected system exit because of bad options")
        except SystemExit:
            pass
        try:
            driver.parse_args(['foo'])
            self.fail("Expected error because bad subcommand is specified")
        except SystemExit:
            pass
        try:
            driver.parse_args(['workflow.selection'])
            self.fail("Expected error because task input is not defined")
        except ValueError:
            pass

    def test_driver2(self):
        driver = pyutilib.workflow.TaskDriver()
        driver.register_task('workflow.selection')
        try:
            # Set sys.argv and then parse arguments
            sys.argv = ['foo', '-h']
            driver.parse_args()
            self.fail("Expected system exit because of bad options")
        except SystemExit:
            pass


if __name__ == "__main__":
    unittest.main()



#
# Unit Tests for yaml_utils
#
#

import os
from os.path import abspath, dirname
currdir = dirname(abspath(__file__)) + os.sep
import pyutilib.th as unittest
import pyutilib.misc

try:
    import yaml
    using_yaml = True
except ImportError:
    using_yaml = False


class Test(unittest.TestCase):

    def setUp(self):
        if not using_yaml:
            self.skipTest("Cannot execute test because YAML is not available.")

    def test1(self):
        # Verify that data 1 is not equal to data 2
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_yaml_files,
            currdir + 'yamldata1.yml',
            currdir + 'yamldata2.yml',
            exact=True)

    def test2(self):
        # Verify that data 1 is a subset of data 2
        pyutilib.misc.compare_yaml_files(
            currdir + 'yamldata1.yml', currdir + 'yamldata2.yml', exact=False)

    def test3(self):
        # Verify that data 2 is not a subset of data 1
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_yaml_files,
            currdir + 'yamldata2.yml',
            currdir + 'yamldata1.yml',
            exact=True)

    def test4(self):
        # Verify check to confirm that YAML data has a different type
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_yaml_files,
            currdir + 'yamldata1.yml',
            currdir + 'yamldata3.yml',
            exact=True)

    def test5(self):
        # Verify that data 1 is not equal to data 1
        pyutilib.misc.compare_yaml_files(
            currdir + 'yamldata1.yml', currdir + 'yamldata1.yml', exact=True)

    def test6(self):
        # Verify that data 1 is a subset of data 4
        pyutilib.misc.compare_yaml_files(
            currdir + 'yamldata1.yml',
            currdir + 'yamldata4.txt',
            output_begin='BEGIN',
            output_end='END',
            exact=False)

    def test7(self):
        # Verify that data 6 is a subset of data 5
        pyutilib.misc.compare_yaml_files(
            currdir + 'yamldata6.yml', currdir + 'yamldata5.yml', exact=False)

    def test8(self):
        # Verify that data 5 is not a subset of data 6
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_yaml_files,
            currdir + 'yamldata5.yml',
            currdir + 'yamldata6.yml',
            exact=False)

    def test9(self):
        # Verify that data 7 is not a subset of data 5
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_yaml_files,
            currdir + 'yamldata7.yml',
            currdir + 'yamldata5.yml',
            exact=False)

    def test10(self):
        # Verify that data 6 is not equal to data 5
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_yaml_files,
            currdir + 'yamldata6.yml',
            currdir + 'yamldata5.yml',
            exact=True)

    def test11(self):
        # Verify that data 8 is not equal to data 5
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_yaml_files,
            currdir + 'yamldata8.yml',
            currdir + 'yamldata5.yml',
            exact=True)

    def test12(self):
        # Verify that data 9 is not a subset of data 1
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_yaml_files,
            currdir + 'yamldata9.yml',
            currdir + 'yamldata1.yml',
            exact=True)

    def test13(self):
        # Verify that data 10 is not a subset of data 1
        self.assertRaises(
            ValueError,
            pyutilib.misc.compare_yaml_files,
            currdir + 'yamldata10.yml',
            currdir + 'yamldata1.yml',
            exact=True)

    def test14(self):
        # Verify that data 10 is a subset of data 1 with a loose tolerance
        pyutilib.misc.compare_yaml_files(
            currdir + 'yamldata10.yml',
            currdir + 'yamldata1.yml',
            tolerance=1.0,
            exact=True)

    def test15a(self):
        # Verify that compressed files can be compared
        pyutilib.misc.compare_yaml_files(
            currdir + 'yamldata14.yml',
            currdir + 'yamldata14.yml.gz',
            exact=True)

    def test15b(self):
        # Verify that compressed files can be compared
        pyutilib.misc.compare_yaml_files(
            currdir + 'yamldata14.yml.gz',
            currdir + 'yamldata14.yml',
            exact=True)


class TestSimple(unittest.TestCase):

    def test1(self):
        # Parse yamldata1.yml
        self.assertEqual({
            'a': {'b': 1,
                  'c': 1.3},
            'd': {'e': 'the rain in spain',
                  'f': 'is mostly on the plain'}
        }, pyutilib.misc.simple_yaml_parser(currdir + 'yamldata1.yml'))

    def test2(self):
        # Parse yamldata2.yml
        self.assertEqual({
            'a': {'b': 1,
                  'c': 1.3},
            'd': {'e': 'the rain in spain',
                  'f': 'is mostly on the plain'},
            'g': 'again'
        }, pyutilib.misc.simple_yaml_parser(currdir + 'yamldata2.yml'))

    def test3(self):
        # Parse yamldata3.yml
        self.assertEqual(
            ['a', 'b', 'c'],
            pyutilib.misc.simple_yaml_parser(currdir + 'yamldata3.yml'))

    def test5(self):
        # Parse yamldata5.yml
        self.assertEqual(
            {'a': ['b', 1, {'c': 'ccc',
                            'e': 'eee',
                            'd': 'ddd'}, 1.3],
             'g': 'again'},
            pyutilib.misc.simple_yaml_parser(currdir + 'yamldata5.yml'))

    def test6(self):
        # Parse yamldata6.yml
        self.assertEqual(
            {'a': [1, {'c': 'ccc',
                       'e': 'eee',
                       'd': 'ddd'}, 1.3],
             'g': 'again'},
            pyutilib.misc.simple_yaml_parser(currdir + 'yamldata6.yml'))

    def test7(self):
        # Parse yamldata7.yml
        self.assertEqual(
            {'a': [1, {'c': 'ccc',
                       'e': 'eee',
                       'd': 'ddd'}, 1.3, 'x', 'y'],
             'g': 'again'},
            pyutilib.misc.simple_yaml_parser(currdir + 'yamldata7.yml'))

    def test11(self):
        # Parse yamldata11.yml
        self.assertEqual(
            ['a', ['b', 'c', ['d', 'e'], 'f', 'g'], 'h'],
            pyutilib.misc.simple_yaml_parser(currdir + 'yamldata11.yml'))

    def test12(self):
        # Parse yamldata12.yml
        self.assertEqual(
            ['a', {'X': ['b', ['d', 'e'], 'f', 'g'],
                   'Y': 'FOO'}, 'h'],
            pyutilib.misc.simple_yaml_parser(currdir + 'yamldata12.yml'))

    def test13(self):
        # Parse yamldata13.yml
        pyutilib.misc.simple_yaml_parser(currdir + 'yamldata13.yml').keys()

    def test14(self):
        # Parse yamldata14.yml
        pyutilib.misc.simple_yaml_parser(currdir + 'yamldata14.yml').keys()


if __name__ == "__main__":
    unittest.main()



#
# Unit Tests for util/data
#
#

import os
from os.path import abspath, dirname, join
pkgdir = dirname(abspath(__file__))

import pyutilib.th as unittest
import pyutilib.excel

try:
    unicode
except:

    def unicode(x):
        return x


try:
    from win32com.client.dynamic import Dispatch
    from pyutilib.excel.spreadsheet_win32com import ExcelSpreadsheet_win32com
    _win32com_available = True  #pragma:nocover
except:
    _win32com_available = False  #pragma:nocover
_excel_available = False  #pragma:nocover
if _win32com_available:
    tmp = ExcelSpreadsheet_win32com()
    try:
        tmp._excel_dispatch()
        tmp._excel_quit()
        _excel_available = True
    except:
        pass
try:
    import xlrd
    _xlrd_available = True
except:
    _xlrd_available = False
try:
    import openpyxl
    _openpyxl_available = True
except:
    _openpyxl_available = False


class BaseTests(object):

    def test_spreadsheet1(self):
        # Create a spreadsheet with a constructor
        sheet = pyutilib.excel.ExcelSpreadsheet(
            join(pkgdir, "test_data" + self.suffix), ctype=self.ctype)
        tmp = sheet.get_range("Arange")

        self.assertEqual(tmp, ['A1', 'A2', 'A3'])
        del sheet

    def test_spreadsheet2(self):
        # Create and open spreadsheet
        sheet = pyutilib.excel.ExcelSpreadsheet(ctype=self.ctype)
        sheet.open(join(pkgdir, "test_data" + self.suffix))
        tmp = sheet.get_range("Arange")
        self.assertEqual(tmp, ['A1', 'A2', 'A3'])
        tmp = sheet.get_range_nrows("Arange")
        self.assertEqual(tmp, 3)
        tmp = sheet.get_range_ncolumns("Arange")
        self.assertEqual(tmp, 1)

    def test_spreadsheet3(self):
        # Create and open spreadsheet in the test dir
        os.chdir(pkgdir)
        sheet = pyutilib.excel.ExcelSpreadsheet(ctype=self.ctype)
        sheet.open("test_data" + self.suffix)
        tmp = sheet.get_range("Arange")
        self.assertEqual(tmp, ['A1', 'A2', 'A3'])
        try:
            tmp = sheet.get_range("Brange")
            self.fail("test_spreadsheet3 - should not have opened range")
        except IOError:
            pass

    def test_spreadsheet4(self):
        # Create and delete spreadsheet without opening it
        sheet = pyutilib.excel.ExcelSpreadsheet(ctype=self.ctype)
        sheet.close()

    def test_spreadsheet5(self):
        # Create and open spreadsheet
        s1 = pyutilib.excel.ExcelSpreadsheet(ctype=self.ctype)
        s1.open(join(pkgdir, "test_data" + self.suffix))
        tmp = s1.get_range("Arange")
        self.assertEqual(tmp, ['A1', 'A2', 'A3'])

    def test_spreadsheet6(self):
        # Create and open spreadsheet
        sheet = pyutilib.excel.ExcelSpreadsheet(ctype=self.ctype)
        if not sheet.can_write():
            self.skipTest("Writing not supported")
        sheet.open(join(pkgdir, "test_data" + self.suffix))
        val = (('B1',), ('B2',), ('B3',))
        sheet.set_range("Arange", val)
        tmp = sheet.get_range("Arange")
        self.assertEqual(tmp, ['B1', 'B2', 'B3'])

    def test_spreadsheet7(self):
        # Create and open spreadsheet
        sheet = pyutilib.excel.ExcelSpreadsheet(ctype=self.ctype)
        if not sheet.can_write():
            self.skipTest("Writing not supported")
        sheet.open(join(pkgdir, "test_data" + self.suffix))
        val = [['B1'], ['B2'], ['B3']]
        sheet.set_range("Arange", val)
        tmp = sheet.get_range("Arange")
        self.assertEqual(tmp, ['B1', 'B2', 'B3'])
        val = [['B1'], ['B2'], ['B3'], ['B4']]
        try:
            sheet.set_range("Arange", val)
            self.fail("expected error")
        except IOError:
            pass

    def test_spreadsheet8(self):
        # Create and open spreadsheet
        sheet = pyutilib.excel.ExcelSpreadsheet(ctype=self.ctype)
        if not sheet.can_write():
            self.skipTest("Writing not supported")
        sheet.open(join(pkgdir, "test_data" + self.suffix))
        val = [['B1', 'C1'], ['B2', 'C1'], ['B3', 'C1']]
        sheet.set_array(3, 4, val)
        val = sheet.get_array(4, 4, 5, 5)
        self.assertEqual(
            val,
            [[unicode('B2'), unicode('C1')], [unicode('B3'), unicode('C1')]])

    def test_spreadsheet9(self):
        # Create and open spreadsheet
        sheet = pyutilib.excel.ExcelSpreadsheet(ctype=self.ctype)
        if not sheet.can_write():
            self.skipTest("Writing not supported")
        sheet.open(join(pkgdir, "test_data" + self.suffix))
        val = (('B1', 'C1'), ('B2', 'C1'), ('B3', 'C1'))
        sheet.set_array(3, 4, val)
        val = sheet.get_array(4, 4, 5, 5)
        self.assertEqual(
            val,
            [[unicode('B2'), unicode('C1')], [unicode('B3'), unicode('C1')]])

    def test_spreadsheet10(self):
        # Verify that we can get updated function values
        sheet = pyutilib.excel.ExcelSpreadsheet(ctype=self.ctype)
        if not sheet.can_calculate():
            self.skipTest("Calculation not supported")
        sheet.open(join(pkgdir, "test_data" + self.suffix))
        sheet.set_range("x", 2.0)
        val = sheet.get_range("x")
        self.assertEqual(val, 2.0)
        val = sheet.get_range("xSquared")
        self.assertEqual(val, 16.0)
        sheet.calculate()
        val = sheet.get_range("xSquared")
        self.assertEqual(val, 4.0)
        sheet.set_range("x", 4.0)
        sheet.close()

    def test_spreadsheet11(self):
        # Verify that we can activate a spreadsheet
        sheet = pyutilib.excel.ExcelSpreadsheet(ctype=self.ctype)
        if not sheet.can_write():
            self.skipTest("Writing not supported")
        sheet.open(join(pkgdir, "test_data" + self.suffix))
        sheet.activate("Sheet2")
        val = sheet.get_range("sInfo")
        self.assertEqual(val, ["s1", "s2", "s3"])
        val = sheet.get_range("A2:A4")
        self.assertEqual(val, ["s1", "s2", "s3"])
        sheet.close()

    def test_calc_iterations(self):
        # Verify that iterations can be set
        sheet = pyutilib.excel.ExcelSpreadsheet(ctype=self.ctype)
        if not sheet.can_calculate():
            self.skipTest("Calculation not supported")
        sheet.open(join(pkgdir, "test_data" + self.suffix))
        tmp = sheet.calc_iterations()
        sheet.calc_iterations(not tmp)
        ttmp = sheet.calc_iterations()
        if tmp is ttmp:
            self.fail("Tried to set calc_iterations to a different value")
        try:
            sheet.calc_iterations(1)
            self.fail("Expected error setting an integer for calc_iterations")
        except ValueError:
            pass
        sheet.close()

    def test_max_iterations(self):
        # Verify that max iterations can be set
        sheet = pyutilib.excel.ExcelSpreadsheet(ctype=self.ctype)
        if not sheet.can_calculate():
            self.skipTest("Calculation not supported")
        sheet.open(join(pkgdir, "test_data" + self.suffix))
        tmp = sheet.max_iterations()
        sheet.max_iterations(max(tmp + 1, 0))
        ttmp = sheet.calc_iterations()
        if tmp is ttmp:
            self.fail("Tried to set max_iterations to a different value")
        try:
            sheet.max_iterations(-1)
            self.fail(
                "Expected error setting a negative integer for max_iterations")
        except ValueError:
            pass
        sheet.close()

    def Xtest_get_column(self):
        """ Verify that we can get a column"""
        sheet = pyutilib.excel.ExcelSpreadsheet(ctype=self.ctype)
        sheet.open(join(pkgdir, "test_data" + self.suffix))
        val = sheet.get_column("E")
        self.assertEqual(val, (1.0, 2.0, None, 4.0, 5.0))
        val = sheet.get_column("F", contiguous=True)
        self.assertEqual(val, (1.0, 2.0, 3.0))
        val = sheet.get_column("G")
        self.assertEqual(val, (1.0, 2.0, 3.0, 4.0, None, 6.0))
        val = sheet.get_column("G", contiguous=True)
        self.assertEqual(val, (1.0, 2.0, 3.0, 4.0))
        sheet.close()

    def Xtest_range1(self):
        """ Verify that getting range will fail if the worksheets are not setup """
        sheet = pyutilib.excel.ExcelSpreadsheet(
            join(pkgdir, "test_data" + self.suffix), ctype=self.ctype)
        try:
            tmp = sheet.get_range("sInfo")
            self.fail(
                "Expected IOError because the range does not exist on the first sheet.")
        except IOError:
            pass
        del sheet

    def Xtest_range2(self):
        """ Verify that ranges can be found on 'other' sheets """
        sheet = pyutilib.excel.ExcelSpreadsheet(
            join(pkgdir, "test_data" + self.suffix), ctype=self.ctype)
        try:
            tmp = sheet.get_range("sInfo")
            self.fail(
                "Expected IOError because the range does not exist on the first sheet.",
                (1, 2))
        except IOError:
            pass
        del sheet


@unittest.skipIf(not _win32com_available, "Cannot import win32com")
@unittest.skipIf(not _excel_available, "Excel not installed")
class Test_win32com(BaseTests, unittest.TestCase):

    ctype = 'win32com'
    suffix = '.xls'


@unittest.skipIf(not _openpyxl_available, "Cannot import openpyxl")
class Test_openpyxl(BaseTests, unittest.TestCase):

    ctype = 'openpyxl'
    suffix = '.xlsx'


@unittest.skipIf(not _xlrd_available, "Cannot import xlrd")
class Test_xlrd(BaseTests, unittest.TestCase):

    ctype = 'xlrd'
    suffix = '.xls'


if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

"""=================================
The PyUtilib Configuration System
=================================

The PyUtilib config system provides a set of three classes
(:py:class:`ConfigDict`, :py:class:`ConfigList`, and
:py:class:`ConfigValue`) for managing and documenting structured
configuration information and user input.  The system is based around
the ConfigValue class, which provides storage for a single configuration
entry.  ConfigValue objects can be grouped using two containers
(ConfigDict and ConfigList), which provide functionality analogous to
Python's dict and list classes, respectively.

At its simplest, the Config system allows for developers to specify a
dictionary of documented configuration entries, allow users to provide
values for those entries, and retrieve the current values:

.. doctest::
    :hide:

    >>> import argparse
    >>> from pyutilib.misc.config import (
    ...     ConfigDict, ConfigList, ConfigValue, In,
    ... )

.. doctest::

    >>> config = ConfigDict()
    >>> config.declare('filename', ConfigValue(
    ...     default=None,
    ...     domain=str,
    ...     description="Input file name",
    ... ))
    <pyutilib.misc.config.ConfigValue object at ...>
    >>> config.declare("bound tolerance", ConfigValue(
    ...     default=1E-5,
    ...     domain=float,
    ...     description="Bound tolerance",
    ...     doc="Relative tolerance for bound feasibility checks"
    ... ))
    <pyutilib.misc.config.ConfigValue object at ...>
    >>> config.declare("iteration limit", ConfigValue(
    ...     default=30,
    ...     domain=int,
    ...     description="Iteration limit",
    ...     doc="Number of maximum iterations in the decomposition methods"
    ... ))
    <pyutilib.misc.config.ConfigValue object at ...>
    >>> config['filename'] = 'tmp.txt'
    >>> print(config['filename'])
    tmp.txt
    >>> print(config['iteration limit'])
    30

For convenience, ConfigDict objects support read/write access via
attributes (with spaces in the declaration names replaced by
underscores):

.. doctest::

    >>> print(config.filename)
    tmp.txt
    >>> print(config.iteration_limit)
    30
    >>> config.iteration_limit = 20
    >>> print(config.iteration_limit)
    20

All Config objects support a ``domain`` keyword that accepts a callable
object (type, function, or callable instance).  The domain callable
should take data and map it onto the desired domain, optionally
performing domain validation (see :py:class:`ConfigValue`,
:py:class:`ConfigDict`, and :py:class:`ConfigList` for more
information).  This allows client code to accept a very flexible set of
inputs without "cluttering" the code with input validation:

.. doctest::

    >>> config.iteration_limit = 35.5
    >>> print(config.iteration_limit)
    35
    >>> print(type(config.iteration_limit).__name__)
    int

Configuring class hierarchies
=============================

A feature of the Config system is that the core classes all implement
``__call__``, and can themselves be used as ``domain`` values.  Beyond
providing domain verification for complex hierarchical structures, this
feature allows ConfigDicts to cleanly support the configuration of
derived objects.  Consider the following example:

.. doctest::

    >>> class Base(object):
    ...     CONFIG = ConfigDict()
    ...     CONFIG.declare('filename', ConfigValue(
    ...         default='input.txt',
    ...         domain=str,
    ...     ))
    ...     def __init__(self, **kwds):
    ...         c = self.CONFIG(kwds)
    ...         c.display()
    ...
    >>> class Derived(Base):
    ...     CONFIG = Base.CONFIG()
    ...     CONFIG.declare('pattern', ConfigValue(
    ...         default=None,
    ...         domain=str,
    ...     ))
    ...
    >>> tmp = Base(filename='foo.txt')
    filename: foo.txt
    >>> tmp = Derived(pattern='.*warning')
    filename: input.txt
    pattern: .*warning

Here, the base class ``Base`` declares a class-level attribute CONFIG as a
ConfigDict containing a single entry (``filename``).  The derived class
(``Derived``) then starts by making a copy of the base class' ``CONFIG``,
and then defines an additional entry (`pattern`).  Instances of the base
class will still create ``c`` instances that only have the single
``filename`` entry, whereas instances of the derived class will have ``c``
instances with two entries: the ``pattern`` entry declared by the derived
class, and the ``filename`` entry "inherited" from the base class.

An extension of this design pattern provides a clean approach for
handling "ephemeral" instance options.  Consider an interface to an
external "solver".  Our class implements a ``solve()`` method that takes a
problem and sends it to the solver along with some solver configuration
options.  We would like to be able to set those options "persistently"
on instances of the interface class, but still override them
"temporarily" for individual calls to ``solve()``.  We implement this by
creating copies of the class's configuration for both specific instances
and for use by each ``solve()`` call:

.. doctest::

    >>> class Solver(object):
    ...     CONFIG = ConfigDict()
    ...     CONFIG.declare('iterlim', ConfigValue(
    ...         default=10,
    ...         domain=int,
    ...     ))
    ...     def __init__(self, **kwds):
    ...         self.config = self.CONFIG(kwds)
    ...     def solve(self, model, **options):
    ...         config = self.config(options)
    ...         # Solve the model with the specified iterlim
    ...         config.display()
    ...
    >>> solver = Solver()
    >>> solver.solve(None)
    iterlim: 10
    >>> solver.config.iterlim = 20
    >>> solver.solve(None)
    iterlim: 20
    >>> solver.solve(None, iterlim=50)
    iterlim: 50
    >>> solver.solve(None)
    iterlim: 20


Interacting with argparse
=========================

In addition to basic storage and retrieval, the Config system provides
hooks to the argparse command-line argument parsing system.  Individual
Config entries can be declared as argparse arguments.  To make
declaration simpler, the :py:meth:`declare` method returns the declared Config
object so that the argument declaration can be done inline:

.. doctest::

    >>> config = ConfigDict()
    >>> config.declare('iterlim', ConfigValue(
    ...     domain=int,
    ...     default=100,
    ...     description="iteration limit",
    ... )).declare_as_argument()
    <pyutilib.misc.config.ConfigValue object at ...>
    >>> config.declare('lbfgs', ConfigValue(
    ...     domain=bool,
    ...     description="use limited memory BFGS update",
    ... )).declare_as_argument()
    <pyutilib.misc.config.ConfigValue object at ...>
    >>> config.declare('linesearch', ConfigValue(
    ...     domain=bool,
    ...     default=True,
    ...     description="use line search",
    ... )).declare_as_argument()
    <pyutilib.misc.config.ConfigValue object at ...>
    >>> config.declare('relative tolerance', ConfigValue(
    ...     domain=float,
    ...     description="relative convergence tolerance",
    ... )).declare_as_argument('--reltol', '-r', group='Tolerances')
    <pyutilib.misc.config.ConfigValue object at ...>
    >>> config.declare('absolute tolerance', ConfigValue(
    ...     domain=float,
    ...     description="absolute convergence tolerance",
    ... )).declare_as_argument('--abstol', '-a', group='Tolerances')
    <pyutilib.misc.config.ConfigValue object at ...>

The ConfigDict can then be used to initialize (or augment) an argparse
ArgumentParser object:

.. doctest::

    >>> parser = argparse.ArgumentParser("tester")
    >>> config.initialize_argparse(parser)


Key information from the ConfigDict is automatically transferred over
to the ArgumentParser object:

.. doctest::

    >>> print(parser.format_help())
    usage: tester [-h] [--iterlim INT] [--lbfgs] [--disable-linesearch]
                  [--reltol FLOAT] [--abstol FLOAT]
    <BLANKLINE>
    optional arguments:
      -h, --help            show this help message and exit
      --iterlim INT         iteration limit
      --lbfgs               use limited memory BFGS update
      --disable-linesearch  [DON'T] use line search
    <BLANKLINE>
    Tolerances:
      --reltol FLOAT, -r FLOAT
                            relative convergence tolerance
      --abstol FLOAT, -a FLOAT
                            absolute convergence tolerance
    <BLANKLINE>

Parsed arguments can then be imported back into the ConfigDict:

.. doctest::

    >>> args=parser.parse_args(['--lbfgs', '--reltol', '0.1', '-a', '0.2'])
    >>> args = config.import_argparse(args)
    >>> config.display()
    iterlim: 100
    lbfgs: true
    linesearch: true
    relative tolerance: 0.1
    absolute tolerance: 0.2

Accessing user-specified values
===============================

It is frequently useful to know which values a user explicitly set, and
which values a user explicitly set, but have never been retrieved.  The
configuration system provides two gemerator methods to return the items
that a user explicitly set (:py:meth:`user_values`) and the items that
were set but never retrieved (:py:meth:`unused_user_values`):

.. doctest::

    >>> print([val.name() for val in config.user_values()])
    ['lbfgs', 'relative tolerance', 'absolute tolerance']
    >>> print(config.relative_tolerance)
    0.1
    >>> print([val.name() for val in config.unused_user_values()])
    ['lbfgs', 'absolute tolerance']

Generating output & documentation
=================================

Configuration objects support three methods for generating output and
documentation: :py:meth:`display()`,
:py:meth:`generate_yaml_template()`, and
:py:meth:`generate_documentation()`.  The simplest is
:py:meth:`display()`, which prints out the current values of the
configuration object (and if it is a container type, all of it's
children).  :py:meth:`generate_yaml_template` is simular to
:py:meth:`display`, but also includes the description fields as
formatted comments.

.. doctest::

    >>> solver_config = config
    >>> config = ConfigDict()
    >>> config.declare('output', ConfigValue(
    ...     default='results.yml',
    ...     domain=str,
    ...     description='output results filename'
    ... ))
    <pyutilib.misc.config.ConfigValue object at ...>
    >>> config.declare('verbose', ConfigValue(
    ...     default=0,
    ...     domain=int,
    ...     description='output verbosity',
    ...     doc='This sets the system verbosity.  The default (0) only logs '
    ...     'warnings and errors.  Larger integer values will produce '
    ...     'additional log messages.',
    ... ))
    <pyutilib.misc.config.ConfigValue object at ...>
    >>> config.declare('solvers', ConfigList(
    ...     domain=solver_config,
    ...     description='list of solvers to apply',
    ... ))
    <pyutilib.misc.config.ConfigList object at ...>
    >>> config.display()
    output: results.yml
    verbose: 0
    solvers: []
    >>> print(config.generate_yaml_template())
    output: results.yml  # output results filename
    verbose: 0           # output verbosity
    solvers: []          # list of solvers to apply
    <BLANKLINE>

It is important to note that both methods document the current state of
the configuration object.  So, in the example above, since the `solvers`
list is empty, you will not get any information on the elements in the
list.  Of course, if you add a value to the list, then the data will be
output:

.. doctest::

    >>> tmp = config()
    >>> tmp.solvers.append({})
    >>> tmp.display()
    output: results.yml
    verbose: 0
    solvers:
      -
        iterlim: 100
        lbfgs: true
        linesearch: true
        relative tolerance: 0.1
        absolute tolerance: 0.2
    >>> print(tmp.generate_yaml_template())
    output: results.yml          # output results filename
    verbose: 0                   # output verbosity
    solvers:                     # list of solvers to apply
      -
        iterlim: 100             # iteration limit
        lbfgs: true              # use limited memory BFGS update
        linesearch: true         # use line search
        relative tolerance: 0.1  # relative convergence tolerance
        absolute tolerance: 0.2  # absolute convergence tolerance
    <BLANKLINE>

The third method (:py:meth:`generate_documentation`) behaves
differently.  This method is designed to generate reference
documentation.  For each configuration item, the `doc` field is output.
If the item has no `doc`, then the `description` field is used.

List containers have their *domain* documented and not their current
values.  The documentation can be configured through optional arguments.
The defaults generate LaTeX documentation:

.. doctest::

    >>> print(config.generate_documentation())
    \\begin{description}[topsep=0pt,parsep=0.5em,itemsep=-0.4em]
      \\item[{output}]\hfill
        \\\\output results filename
      \\item[{verbose}]\hfill
        \\\\This sets the system verbosity.  The default (0) only logs warnings and
        errors.  Larger integer values will produce additional log messages.
      \\item[{solvers}]\hfill
        \\\\list of solvers to apply
      \\begin{description}[topsep=0pt,parsep=0.5em,itemsep=-0.4em]
        \\item[{iterlim}]\hfill
          \\\\iteration limit
        \\item[{lbfgs}]\hfill
          \\\\use limited memory BFGS update
        \\item[{linesearch}]\hfill
          \\\\use line search
        \\item[{relative tolerance}]\hfill
          \\\\relative convergence tolerance
        \\item[{absolute tolerance}]\hfill
          \\\\absolute convergence tolerance
      \\end{description}
    \\end{description}
    <BLANKLINE>

"""

import re
import sys
from textwrap import wrap
import logging
import pickle

import six
from six.moves import xrange

def _dump(*args, **kwds):
    try:
        from yaml import dump
    except ImportError:
        #dump = lambda x,**y: str(x)
        # YAML uses lowercase True/False
        def dump(x, **args):
            if type(x) is bool:
                return str(x).lower()
            return str(x)
    assert '_dump' in globals()
    globals()['_dump'] = dump
    return dump(*args, **kwds)

try:
    import argparse
    argparse_is_available = True
except ImportError:
    argparse_is_available = False

try:
    import builtins as _builtins
except ImportError:
    import __builtin__ as _builtins

__all__ = ('ConfigDict', 'ConfigBlock', 'ConfigList', 'ConfigValue')

logger = logging.getLogger('pyutilib.misc.config')


def _munge_name(name, space_to_dash=True):
    if space_to_dash:
        name = re.sub(r'\s', '-', name)
    name = re.sub(r'_', '-', name)
    return re.sub(r'[^a-zA-Z0-9-_]', '_', name)


_leadingSpace = re.compile('^([ \n\t]*)')

def _strip_indentation(doc):
    if not doc:
        return doc
    lines = doc.splitlines()
    while lines and not lines[0].strip():
        lines.pop(0)
    if len(lines) == 1:
        return doc.lstrip()
    minIndent = min(len(_leadingSpace.match(l).group(0)) for l in lines[1:])
    if len(_leadingSpace.match(lines[0]).group(0)) <= minIndent:
        lines[0] = lines[0].strip()
    else:
        lines[0] = lines[0][minIndent:].rstrip()
    for i, l in enumerate(lines[1:]):
        lines[i + 1] = l[minIndent:].rstrip()
    return '\n'.join(lines)


def _value2string(prefix, value, obj):
    _str = prefix
    if value is not None:
        try:
            _data = value._data if value is obj else value
            if getattr(_builtins, _data.__class__.__name__, None
                   ) is not None:
                _str += _dump(_data, default_flow_style=True).rstrip()
                if _str.endswith("..."):
                    _str = _str[:-3].rstrip()
            else:
                _str += str(_data)
        except:
            _str += str(type(_data))
    return _str.rstrip()

def _value2yaml(prefix, value, obj):
    _str = prefix
    if value is not None:
        try:
            _data = value._data if value is obj else value
            _str += _dump(_data, default_flow_style=True).rstrip()
            if _str.endswith("..."):
                _str = _str[:-3].rstrip()
        except:
            _str += str(type(_data))
    return _str.rstrip()


class _UnpickleableDomain(object):
    def __init__(self, obj):
        self._type = type(obj).__name__
        self._name = obj.name(True)

    def __call__(self, arg):
        logging.error(
"""%s %s was pickled with an unpicklable domain.
    The domain was stripped and lost during the pickle process.  Setting
    new values on the restored object cannot be mapped into the correct
    domain.
""" % ( self._type, self._name))
        return arg

def _picklable(field,obj):
    ftype = type(field)
    if ftype in _picklable.known:
        return field if _picklable.known[ftype] else _UnpickleableDomain(obj)
    try:
        pickle.dumps(field)
        _picklable.known[ftype] = True
        return field
    except:
        # Contrary to the documentation, Python is not at all consistent
        # with the exception that is raised when pickling an object
        # fails:
        #
        #    Python 2.6 - 3.4:  pickle.PicklingError
        #    Python 3.5 - 3.6:  AttributeError
        #    Python 2.6 - 2.7 (cPickle):  TypeError
        #
        # What we are concerned about is masking things like recursion
        # errors.  Unfortunately, Python is not quite consistent there,
        # either: exceeding recursion depth raises a RuntimeError
        # through 3.4, then switches to a RecursionError (a derivative
        # of RuntimeError).
        if isinstance(sys.exc_info()[0], RuntimeError):
            raise
        _picklable.known[ftype] = False
        return _UnpickleableDomain(obj)

_picklable.known = {}

class ConfigBase(object):
    __slots__ = ('_parent', '_name', '_userSet', '_userAccessed', '_data',
                 '_default', '_domain', '_description', '_doc', '_visibility',
                 '_argparse')

    # This just needs to be any singleton-like object; we use it so that
    # we can tell if an argument is provided (and we can't use None as
    # None is a valid user-specified argument).  Making it a class helps
    # when Config objects are pickled.
    class NoArgument(object): pass

    def __init__(self,
                 default=None,
                 domain=None,
                 description=None,
                 doc=None,
                 visibility=0):
        self._parent = None
        self._name = None
        self._userSet = False
        self._userAccessed = False

        self._data = None
        self._default = default
        self._domain = domain
        self._description = _strip_indentation(description)
        self._doc = _strip_indentation(doc)
        self._visibility = visibility
        self._argparse = None

    def __getstate__(self):
        # Nominally, __getstate__() should return:
        #
        # state = super(Class, self).__getstate__()
        # for i in Class.__slots__:
        #    state[i] = getattr(self,i)
        # return state
        #
        # Hoewever, in this case, the (nominal) parent class is
        # 'object', and object does not implement __getstate__.  Since
        # super() doesn't actually return a class, we are going to check
        # the *derived class*'s MRO and see if this is the second to
        # last class (the last is always 'object').  If it is, then we
        # can allocate the state dictionary.  If it is not, then we call
        # the super-class's __getstate__ (since that class is NOT
        # 'object').
        if self.__class__.__mro__[-2] is ConfigBase:
            state = {}
        else:
            state = super(ConfigBase, self).__getstate__()
        state.update((key, getattr(self, key)) for key in ConfigBase.__slots__)
        state['_domain'] = _picklable(state['_domain'], self)
        state['_parent'] = None
        return state

    def __setstate__(self, state):
        for key, val in six.iteritems(state):
            # Note: per the Python data model docs, we explicitly
            # set the attribute using object.__setattr__() instead
            # of setting self.__dict__[key] = val.
            object.__setattr__(self, key, val)

    def __call__(self, value=NoArgument, default=NoArgument, domain=NoArgument,
                 description=NoArgument, doc=NoArgument, visibility=NoArgument,
                 implicit=NoArgument, implicit_domain=NoArgument,
                 preserve_implicit=False):
        # We will pass through overriding arguments to the constructor.
        # This way if the constructor does special processing of any of
        # the arguments (like implicit_domain), we don't have to repeat
        # that code here.  Unfortunately, it means we need to do a bit
        # of logic to be sure we only pass through appropriate
        # arguments.
        kwds = {}
        kwds['description'] = ( self._description
                                if description is ConfigBase.NoArgument else
                                description )
        kwds['doc'] = ( self._doc
                        if doc is ConfigBase.NoArgument else
                        doc )
        kwds['visibility'] = ( self._visibility
                               if visibility is ConfigBase.NoArgument else
                               visibility )
        if isinstance(self, ConfigDict):
            kwds['implicit'] = ( self._implicit_declaration
                                 if implicit is ConfigBase.NoArgument else
                                 implicit )
            kwds['implicit_domain'] = (
                self._implicit_domain
                if implicit_domain is ConfigBase.NoArgument else
                implicit_domain )
            if domain is not ConfigBase.NoArgument:
                logger.warn("domain ignored by __call__(): "
                            "class is a ConfigDict" % (type(self),))
            if default is not ConfigBase.NoArgument:
                logger.warn("default ignored by __call__(): "
                            "class is a ConfigDict" % (type(self),))
        else:
            kwds['default'] = ( self.value()
                                if default is ConfigBase.NoArgument else
                                default )
            kwds['domain'] = ( self._domain
                               if domain is ConfigBase.NoArgument else
                               domain )
            if implicit is not ConfigBase.NoArgument:
                logger.warn("implicit ignored by __call__(): "
                            "class %s is not a ConfigDict" % (type(self),))
            if implicit_domain is not ConfigBase.NoArgument:
                logger.warn("implicit_domain ignored by __call__(): "
                            "class %s is not a ConfigDict" % (type(self),))

        # Copy over any other object-specific information (mostly Block
        # definitions)
        ans = self.__class__(**kwds)
        if isinstance(self, ConfigDict):
            for k in self._decl_order:
                if preserve_implicit or k in self._declared:
                    v = self._data[k]
                    ans._data[k] = _tmp = v(preserve_implicit=preserve_implicit)
                    ans._decl_order.append(k)
                    if k in self._declared:
                        ans._declared.add(k)
                    _tmp._parent = ans
                    _tmp._name = v._name
        else:
            ans.reset()
        # ... and set the value, if appropriate
        if value is not ConfigBase.NoArgument:
            ans.set_value(value)
        return ans

    def name(self, fully_qualified=False):
        # Special case for the top-level block
        if self._name is None:
            return ""
        elif fully_qualified and self._parent is not None:
            pName = self._parent.name(fully_qualified)
            # Special case for ConfigList indexing and the top-level entries
            if self._name.startswith('[') or not pName:
                return pName + self._name
            else:
                return pName + '.' + self._name
        else:
            return self._name

    def set_default_value(self, default):
        self._default = default

    def set_domain(self, domain):
        self._domain = domain
        self.set_value(self.value(accessValue=False))

    def _cast(self, value):
        if value is None:
            return value
        if self._domain is not None:
            try:
                if value is not ConfigBase.NoArgument:
                    return self._domain(value)
                else:
                    return self._domain()
            except:
                err = sys.exc_info()[1]
                if hasattr(self._domain, '__name__'):
                    _dom = self._domain.__name__
                else:
                    _dom = type(self._domain)
                raise ValueError("invalid value for configuration '%s':\n"
                                 "\tFailed casting %s\n\tto %s\n\tError: %s" %
                                 (self.name(True), value, _dom, err))
        else:
            return value

    def reset(self):
        #
        # This is a dangerous construct, the failure in the first try block
        # can mask a real problem.
        #
        try:
            self.set_value(self._default)
        except:
            if hasattr(self._default, '__call__'):
                self.set_value(self._default())
            else:
                raise
        self._userAccessed = False
        self._userSet = False

    def declare_as_argument(self, *args, **kwds):
        """Map this Config item to an argparse argument.

        Valid arguments include all valid arguments to argparse's
        ArgumentParser.add_argument() with the exception of 'default'.
        In addition, you may provide a group keyword argument can be
        used to either pass in a pre-defined option group or subparser,
        or else pass in the title of a group, subparser, or (subparser,
        group).

        """

        if 'default' in kwds:
            raise TypeError(
                "You cannot specify an argparse default value with "
                "ConfigBase.declare_as_argument().  The default value is "
                "supplied automatically from the Config definition.")

        if 'action' not in kwds and self._domain is bool:
            if not self._default:
                kwds['action'] = 'store_true'
            else:
                kwds['action'] = 'store_false'
                if not args:
                    args = ('--disable-' + _munge_name(self.name()),)
                if 'help' not in kwds:
                    kwds['help'] = "[DON'T] " + self._description
        if 'help' not in kwds:
            kwds['help'] = self._description
        if not args:
            args = ('--' + _munge_name(self.name()),)
        if self._argparse:
            self._argparse = self._argparse + ((args, kwds),)
        else:
            self._argparse = ((args, kwds),)
        return self

    def initialize_argparse(self, parser):

        def _get_subparser_or_group(_parser, name):
            # Note: strings also have a 'title()' method.  We are
            # looking for things that look like argparse
            # groups/subparsers, so just checking for the attribute
            # is insufficient: it needs to be a string attribute as
            # well
            if isinstance(name, argparse._ActionsContainer):
                #hasattr(_group, 'title') and \
                #    isinstance(_group.title, six.string_types):
                return 2, name

            if not isinstance(name, six.string_types):
                raise RuntimeError(
                    'Unknown datatype (%s) for argparse group on '
                    'configuration definition %s' %
                    (type(name).__name__, obj.name(True)))

            try:
                for _grp in _parser._subparsers._group_actions:
                    if name in _grp._name_parser_map:
                        return 1, _grp._name_parser_map[name]
            except AttributeError:
                pass

            for _grp in _parser._action_groups:
                if _grp.title == name:
                    return 0, _grp
            return 0, _parser.add_argument_group(title=name)

        def _process_argparse_def(_args, _kwds):
            _parser = parser
            # shallow copy the dict so we can remove the group flag and
            # add things like documentation, etc.
            _kwds = dict(_kwds)
            if 'group' in _kwds:
                _group = _kwds.pop('group')
                if isinstance(_group, tuple):
                    for _idx, _grp in enumerate(_group):
                        _issub, _parser = _get_subparser_or_group(_parser, _grp)
                        if not _issub and _idx < len(_group) - 1:
                            raise RuntimeError(
                                "Could not find argparse subparser '%s' for "
                                "Config item %s" % (_grp, obj.name(True)))
                else:
                    _issub, _parser = _get_subparser_or_group(_parser, _group)
            if 'dest' not in _kwds:
                _kwds['dest'] = 'CONFIGBLOCK.' + obj.name(True)
                if 'metavar' not in _kwds and \
                   _kwds.get('action','') not in ('store_true','store_false'):
                    if obj._domain is not None and \
                       obj._domain.__class__ is type:
                        _kwds['metavar'] = obj._domain.__name__.upper()
                    else:
                        _kwds['metavar'] = _munge_name(self.name().upper(),
                                                       False)
            _parser.add_argument(*_args, default=argparse.SUPPRESS, **_kwds)

        assert (argparse_is_available)
        for level, prefix, value, obj in self._data_collector(None, ""):
            if obj._argparse is None:
                continue
            for _args, _kwds in obj._argparse:
                _process_argparse_def(_args, _kwds)

    def import_argparse(self, parsed_args):
        for level, prefix, value, obj in self._data_collector(None, ""):
            if obj._argparse is None:
                continue
            for _args, _kwds in obj._argparse:
                if 'dest' in _kwds:
                    _dest = _kwds['dest']
                    if _dest in parsed_args:
                        obj.set_value(parsed_args.__dict__[_dest])
                else:
                    _dest = 'CONFIGBLOCK.' + obj.name(True)
                    if _dest in parsed_args:
                        obj.set_value(parsed_args.__dict__[_dest])
                        del parsed_args.__dict__[_dest]
        return parsed_args

    def display(self, content_filter=None, indent_spacing=2, ostream=None,
                visibility=None):
        if content_filter not in ConfigDict.content_filters:
            raise ValueError("unknown content filter '%s'; valid values are %s"
                             % (content_filter, ConfigDict.content_filters))

        _blocks = []
        if ostream is None:
            ostream=sys.stdout

        for lvl, prefix, value, obj in self._data_collector(0, "", visibility):
            if content_filter == 'userdata' and not obj._userSet:
                continue

            _str = _value2string(prefix, value, obj)
            _blocks[lvl:] = [' ' * indent_spacing * lvl + _str + "\n",]

            for i, v in enumerate(_blocks):
                if v is not None:
                    ostream.write(v)
                    _blocks[i] = None

    def generate_yaml_template(self, indent_spacing=2, width=78, visibility=0):
        minDocWidth = 20
        comment = "  # "
        data = list(self._data_collector(0, "", visibility))
        level_info = {}
        for lvl, pre, val, obj in data:
            _str = _value2yaml(pre, val, obj)
            if lvl not in level_info:
                level_info[lvl] = {'data': [], 'off': 0, 'line': 0, 'over': 0}
            level_info[lvl]['data'].append(
                (_str.find(':') + 2, len(_str), len(obj._description or "")))
        for lvl in sorted(level_info):
            indent = lvl * indent_spacing
            _ok = width - indent - len(comment) - minDocWidth
            offset = \
                max( val if val < _ok else key
                     for key,val,doc in level_info[lvl]['data'] )
            offset += indent + len(comment)
            over = sum(1 for key, val, doc in level_info[lvl]['data']
                       if doc + offset > width)
            if len(level_info[lvl]['data']) - over > 0:
                line = max(offset + doc
                           for key, val, doc in level_info[lvl]['data']
                           if offset + doc <= width)
            else:
                line = width
            level_info[lvl]['off'] = offset
            level_info[lvl]['line'] = line
            level_info[lvl]['over'] = over
        maxLvl = 0
        maxDoc = 0
        pad = 0
        for lvl in sorted(level_info):
            _pad = level_info[lvl]['off']
            _doc = level_info[lvl]['line'] - _pad
            if _pad > pad:
                if maxDoc + _pad <= width:
                    pad = _pad
                else:
                    break
            if _doc + pad > width:
                break
            if _doc > maxDoc:
                maxDoc = _doc
            maxLvl = lvl
        os = six.StringIO()
        if self._description:
            os.write(comment.lstrip() + self._description + "\n")
        for lvl, pre, val, obj in data:
            _str = _value2yaml(pre, val, obj)
            if not obj._description:
                os.write(' ' * indent_spacing * lvl + _str + "\n")
                continue
            if lvl <= maxLvl:
                field = pad - len(comment)
            else:
                field = level_info[lvl]['off'] - len(comment)
            os.write(' ' * indent_spacing * lvl)
            if width - len(_str) - minDocWidth >= 0:
                os.write('%%-%ds' % (field - indent_spacing * lvl) % _str)
            else:
                os.write(_str + '\n' + ' ' * field)
            os.write(comment)
            txtArea = max(width - field - len(comment), minDocWidth)
            os.write(("\n" + ' ' * field + comment).join(
                wrap(
                    obj._description, txtArea, subsequent_indent='  ')))
            os.write('\n')
        return os.getvalue()

    def generate_documentation\
            ( self,
              block_start= "\\begin{description}[topsep=0pt,parsep=0.5em,itemsep=-0.4em]\n",
              block_end=   "\\end{description}\n",
              item_start=  "\\item[{%s}]\\hfill\n",
              item_body=   "\\\\%s",
              item_end=    "",
              indent_spacing=2,
              width=78,
              visibility=0
              ):
        os = six.StringIO()
        level = []
        lastObj = self
        indent = ''
        for lvl, pre, val, obj in self._data_collector(1, '', visibility, True):
            #print len(level), lvl, val, obj
            if len(level) < lvl:
                while len(level) < lvl - 1:
                    level.append(None)
                level.append(lastObj)
                if '%s' in block_start:
                    os.write(indent + block_start % lastObj.name())
                elif block_start:
                    os.write(indent + block_start)
                indent += ' ' * indent_spacing
            while len(level) > lvl:
                _last = level.pop()
                if _last is not None:
                    indent = indent[:-1 * indent_spacing]
                    if '%s' in block_end:
                        os.write(indent + block_end % _last.name())
                    elif block_end:
                        os.write(indent + block_end)

            lastObj = obj
            if '%s' in item_start:
                os.write(indent + item_start % obj.name())
            elif item_start:
                os.write(indent + item_start)
            _doc = obj._doc or obj._description or ""
            if _doc:
                _wrapLines = '\n ' not in _doc
                if '%s' in item_body:
                    _doc = item_body % (_doc,)
                elif _doc:
                    _doc = item_body
                if _wrapLines:
                    doc_lines = wrap(
                        _doc,
                        width,
                        initial_indent=indent + ' ' * indent_spacing,
                        subsequent_indent=indent + ' ' * indent_spacing)
                else:
                    doc_lines = (_doc,)
                # Write things out
                os.writelines('\n'.join(doc_lines))
                if not doc_lines[-1].endswith("\n"):
                    os.write('\n')
            if '%s' in item_end:
                os.write(indent + item_end % obj.name())
            elif item_end:
                os.write(indent + item_end)
        while level:
            _last = level.pop()
            if _last is not None:
                indent = indent[:-1 * indent_spacing]
                if '%s' in block_end:
                    os.write(indent + block_end % _last.name())
                else:
                    os.write(indent + block_end)
        return os.getvalue()

    def user_values(self):
        if self._userSet:
            yield self
        for level, prefix, value, obj in self._data_collector(0, ""):
            if obj._userSet:
                yield obj

    def unused_user_values(self):
        if self._userSet and not self._userAccessed:
            yield self
        for level, prefix, value, obj in self._data_collector(0, ""):
            if obj._userSet and not obj._userAccessed:
                yield obj


class ConfigValue(ConfigBase):
    """Store and manipulate a single configuration value.

    Parameters
    ----------
    default: optional
        The default value that this ConfigValue will take if no value is
        provided.

    domain: callable, optional
        The domain can be any callable that accepts a candidate value
        and returns the value converted to the desired type, optionally
        performing any data validation.  The result will be stored into
        the ConfigValue.  Examples include type constructors like `int`
        or `float`.  More complex domain examples include callable
        objects; for example, the :py:class:`In` class that ensures that
        the value falls into an acceptable set or even a complete
        :py:class:`ConfigDict` instance.

    description: str, optional
        The short description of this value

    doc: str, optional
        The long documentation string for this value

    visibility: int, optional
        The visibility of this ConfigValue when generating templates and
        documentation.  Visibility supports specification of "advanced"
        or "developer" options.  ConfigValues with visibility=0 (the
        default) will always be printed / included.  ConfigValues
        with higher visibility values will only be included when the
        generation method specifies a visibility greater than or equal
        to the visibility of this object.

    """

    def __init__(self, *args, **kwds):
        ConfigBase.__init__(self, *args, **kwds)
        self.reset()

    def value(self, accessValue=True):
        if accessValue:
            self._userAccessed = True
        return self._data

    def set_value(self, value):
        self._data = self._cast(value)
        self._userSet = True

    def _data_collector(self, level, prefix, visibility=None, docMode=False):
        if visibility is not None and visibility < self._visibility:
            return
        yield (level, prefix, self, self)


class ImmutableConfigValue(ConfigValue):
    def set_value(self, value):
        if self._cast(value) != self._data:
            raise RuntimeError(str(self) + ' is currently immutable')
        super(ImmutableConfigValue, self).set_value(value)

    def reset(self):
        try:
            super(ImmutableConfigValue, self).set_value(self._default)
        except:
            if hasattr(self._default, '__call__'):
                super(ImmutableConfigValue, self).set_value(self._default())
            else:
                raise
        self._userAccessed = False
        self._userSet = False


class MarkImmutable(object):
    """
    Mark instances of ConfigValue as immutable.

    Parameters
    ----------
    config_value: ConfigValue
        The ConfigValue instances that should be marked immutable. 
        Note that multiple instances of ConfigValue can be passed.

    Examples
    --------
    >>> config = ConfigBlock()
    >>> config.declare('a', ConfigValue(default=1, domain=int))
    >>> config.declare('b', ConfigValue(default=1, domain=int))
    >>> locker = MarkImmutable(config.get('a'), config.get('b'))
    
    Now, config.a and config.b cannot be changed. To make them mutable again, 

    >>> locker.release_lock()
    """
    def __init__(self, *args):
        self._locked = list()
        try:
            for arg in args:
                if type(arg) is not ConfigValue:
                    raise ValueError('Only ConfigValue instances can be marked immutable.')
                arg.__class__ = ImmutableConfigValue
                self._locked.append(arg)
        except:
            self.release_lock()
            raise

    def release_lock(self):
        for arg in self._locked:
            arg.__class__ = ConfigValue
        self._locked = list()


class ConfigList(ConfigBase):
    """Store and manipulate a list of configuration values.

    Parameters
    ----------
    default: optional
        The default value that this ConfigList will take if no value is
        provided.  If default is a list or ConfigList, then each member
        is cast to the ConfigList's domain to build the default value,
        otherwise the default is cast to the domain and forms a default
        list with a single element.

    domain: callable, optional
        The domain can be any callable that accepts a candidate value
        and returns the value converted to the desired type, optionally
        performing any data validation.  The result will be stored /
        added to the ConfigList.  Examples include type constructors
        like `int` or `float`.  More complex domain examples include
        callable objects; for example, the :py:class:`In` class that
        ensures that the value falls into an acceptable set or even a
        complete :py:class:`ConfigDict` instance.

    description: str, optional
        The short description of this list

    doc: str, optional
        The long documentation string for this list

    visibility: int, optional
        The visibility of this ConfigList when generating templates and
        documentation.  Visibility supports specification of "advanced"
        or "developer" options.  ConfigLists with visibility=0 (the
        default) will always be printed / included.  ConfigLists
        with higher visibility values will only be included when the
        generation method specifies a visibility greater than or equal
        to the visibility of this object.

    """

    def __init__(self, *args, **kwds):
        ConfigBase.__init__(self, *args, **kwds)
        if self._domain is None:
            self._domain = ConfigValue()
        elif isinstance(self._domain, ConfigBase):
            pass
        else:
            self._domain = ConfigValue(None, domain=self._domain)
        self.reset()


    def __setstate__(self, state):
        state = super(ConfigList, self).__setstate__(state)
        for x in self._data:
            x._parent = self

    def __getitem__(self, key):
        val = self._data[key]
        self._userAccessed = True
        if isinstance(val, ConfigValue):
            return val.value()
        else:
            return val

    def get(self, key, default=ConfigBase.NoArgument):
        # Note: get() is borrowed from ConfigDict for cases where we
        # want the raw stored object (and to aviod the implicit
        # conversion of ConfigValue members to their stored data).
        try:
            val = self._data[key]
            self._userAccessed = True
            return val
        except:
            pass
        if default is ConfigBase.NoArgument:
            return None
        if self._domain is not None:
            return self._domain(default)
        else:
            return ConfigValue(default)

    def __setitem__(self, key, val):
        # Note: this will fail if the element doesn't exist in _data.
        # As a result, *this* list doesn't change when someone tries to
        # change an element; instead, the *element* gets its _userSet
        # flag set.
        #self._userSet = True
        self._data[key].set_value(val)

    def __len__(self):
        return self._data.__len__()

    def __iter__(self):
        self._userAccessed = True
        return iter(self[i] for i in xrange(len(self._data)))

    def value(self, accessValue=True):
        if accessValue:
            self._userAccessed = True
        return [config.value(accessValue) for config in self._data]

    def set_value(self, value):
        # If the set_value fails part-way through the list values, we
        # want to restore a deterministic state.  That is, either
        # set_value succeeds completely, or else nothing happens.
        _old = self._data
        self._data = []
        try:
            if (type(value) is list) or \
               isinstance(value, ConfigList):
                for val in value:
                    self.append(val)
            else:
                self.append(value)
        except:
            self._data = _old
            raise
        self._userSet = True

    def reset(self):
        ConfigBase.reset(self)
        # Because the base reset() calls set_value, any deefault list
        # entries will get their userSet flag set.  This is wrong, as
        # reset() should conceptually reset teh object to it's default
        # state (e.g., before the user ever had a chance to mess with
        # things).  As the list could contain a ConfigDict, this is a
        # recursive operation to put the userSet values back.
        for val in self.user_values():
            val._userSet = False

    def append(self, value=ConfigBase.NoArgument):
        val = self._cast(value)
        if val is None:
            return
        self._data.append(val)
        #print self._data[-1], type(self._data[-1])
        self._data[-1]._parent = self
        self._data[-1]._name = '[%s]' % (len(self._data) - 1,)
        self._data[-1]._userSet = True
        self._userSet = True

    def add(self, value=ConfigBase.NoArgument):
        logger.warning(
            "DEPRECATED: ConfigList.add() has been deprecated.  Use append()")
        return self.append(value)

    def _data_collector(self, level, prefix, visibility=None, docMode=False):
        if visibility is not None and visibility < self._visibility:
            return
        if docMode:
            # In documentation mode, we do NOT list the documentation
            # for any sub-data, and instead document the *domain*
            # information (as all the entries should share the same
            # domain, potentially duplicating that documentation is
            # somewhat redundant, and worse, if the list is empty, then
            # no documentation is generated at all!)
            yield (level, prefix, None, self)
            subDomain = self._domain._data_collector(level + 1, '- ',
                                                     visibility, docMode)
            # Pop off the (empty) block entry
            six.next(subDomain)
            for v in subDomain:
                yield v
            return
        if prefix:
            if not self._data:
                yield (level, prefix, [], self)
            else:
                yield (level, prefix, None, self)
                if level is not None:
                    level += 1
        for value in self._data:
            for v in value._data_collector(level, '- ', visibility, docMode):
                yield v


class ConfigDict(ConfigBase):
    """Store and manipulate a dictionary of configuration values.

    Parameters
    ----------
    description: str, optional
        The short description of this list

    doc: str, optional
        The long documentation string for this list

    implicit: bool, optional
        If True, the ConfigDict will allow "implicitly" declared
        keys, that is, keys can be stored into the ConfigDict that
        were not prevously declared using :py:meth:`declare` or
        :py:meth:`declare_from`.

    implicit_domain: callable, optional
        The domain that will be used for any implicitly-declared keys.
        Follows the same rules as :py:meth:`ConfigValue`'s `domain`.

    visibility: int, optional
        The visibility of this ConfigDict when generating templates and
        documentation.  Visibility supports specification of "advanced"
        or "developer" options.  ConfigDicts with visibility=0 (the
        default) will always be printed / included.  ConfigDicts
        with higher visibility values will only be included when the
        generation method specifies a visibility greater than or equal
        to the visibility of this object.

    """

    content_filters = (None, 'all', 'userdata')

    __slots__ = ('_decl_order', '_declared', '_implicit_declaration',
                 '_implicit_domain')
    _all_slots = __slots__ + ConfigBase.__slots__

    def __init__(self,
                 description=None,
                 doc=None,
                 implicit=False,
                 implicit_domain=None,
                 visibility=0):
        self._decl_order = []
        self._declared = set()
        self._implicit_declaration = implicit
        if implicit_domain is None or isinstance(implicit_domain, ConfigBase):
            self._implicit_domain = implicit_domain
        else:
            self._implicit_domain = ConfigValue(None, domain=implicit_domain)
        ConfigBase.__init__(self, None, {}, description, doc, visibility)
        self._data = {}

    def __getstate__(self):
        state = super(ConfigDict, self).__getstate__()
        state.update((key, getattr(self, key)) for key in ConfigDict.__slots__)
        state['_implicit_domain'] = _picklable(state['_implicit_domain'], self)
        return state

    def __setstate__(self, state):
        state = super(ConfigDict, self).__setstate__(state)
        for x in six.itervalues(self._data):
            x._parent = self

    def __getitem__(self, key):
        self._userAccessed = True
        key = str(key)
        if isinstance(self._data[key], ConfigValue):
            return self._data[key].value()
        else:
            return self._data[key]

    def get(self, key, default=ConfigBase.NoArgument):
        self._userAccessed = True
        key = str(key)
        if key in self._data:
            return self._data[key]
        if default is ConfigBase.NoArgument:
            return None
        if self._implicit_domain is not None:
            return self._implicit_domain(default)
        else:
            return ConfigValue(default)

    def setdefault(self, key, default=ConfigBase.NoArgument):
        self._userAccessed = True
        key = str(key)
        if key in self._data:
            return self._data[key]
        if default is ConfigBase.NoArgument:
            return self.add(key, None)
        else:
            return self.add(key, default)

    def __setitem__(self, key, val):
        key = str(key)
        if key not in self._data:
            self.add(key, val)
        else:
            self._data[key].set_value(val)
        #self._userAccessed = True

    def __delitem__(self, key):
        # Note that this will produce a KeyError if the key is not valid
        # for this ConfigDict.
        del self._data[key]
        # Clean up the other data structures
        self._decl_order.remove(key)
        self._declared.discard(key)

    def __contains__(self, key):
        key = str(key)
        return key in self._data

    def __len__(self):
        return self._decl_order.__len__()

    def __iter__(self):
        return self._decl_order.__iter__()

    def __getattr__(self, name):
        # Note: __getattr__ is only called after all "usual" attribute
        # lookup methods have failed.  So, if we get here, we already
        # know that key is not a __slot__ or a method, etc...
        #if name in ConfigDict._all_slots:
        #    return super(ConfigDict,self).__getattribute__(name)
        if name not in self._data:
            _name = name.replace('_', ' ')
            if _name not in self._data:
                raise AttributeError("Unknown attribute '%s'" % name)
            name = _name
        return ConfigDict.__getitem__(self, name)

    def __setattr__(self, name, value):
        if name in ConfigDict._all_slots:
            super(ConfigDict, self).__setattr__(name, value)
        else:
            if name not in self._data:
                name = name.replace('_', ' ')
            ConfigDict.__setitem__(self, name, value)

    def iterkeys(self):
        return self._decl_order.__iter__()

    def itervalues(self):
        self._userAccessed = True
        for key in self._decl_order:
            yield self[key]

    def iteritems(self):
        self._userAccessed = True
        for key in self._decl_order:
            yield (key, self[key])

    def keys(self):
        return list(self.iterkeys())

    def values(self):
        return list(self.itervalues())

    def items(self):
        return list(self.iteritems())

    def _add(self, name, config):
        name = str(name)
        if config._parent is not None:
            raise ValueError(
                "config '%s' is already assigned to Config Block '%s'; "
                "cannot reassign to '%s'" %
                (name, config._parent.name(True), self.name(True)))
        if name in self._data:
            raise ValueError(
                "duplicate config '%s' defined for Config Block '%s'" %
                (name, self.name(True)))
        if '.' in name or '[' in name or ']' in name:
            raise ValueError(
                "Illegal character in config '%s' for config Block '%s': "
                "'.[]' are not allowed." % (name, self.name(True)))
        self._data[name] = config
        self._decl_order.append(name)
        config._parent = self
        config._name = name
        return config

    def declare(self, name, config):
        ans = self._add(name, config)
        self._declared.add(name)
        return ans

    def declare_from(self, other, skip=None):
        if not isinstance(other, ConfigDict):
            raise ValueError(
                "ConfigDict.declare_from() only accepts other ConfigDicts")
        # Note that we duplicate ["other()"] other so that this
        # ConfigDict's entries are independent of the other's
        for key in other.iterkeys():
            if skip and key in skip:
                continue
            if key in self:
                raise ValueError("ConfigDict.declare_from passed a block "
                                 "with a duplicate field, %s" % (key,))
            self.declare(key, other._data[key]())

    def add(self, name, config):
        if not self._implicit_declaration:
            raise ValueError("Key '%s' not defined in Config Block '%s'"
                             " and Block disallows implicit entries" %
                             (name, self.name(True)))

        if self._implicit_domain is None:
            if isinstance(config, ConfigBase):
                ans = self._add(name, config)
            else:
                ans = self._add(name, ConfigValue(config))
        else:
            ans = self._add(name, self._implicit_domain(config))
        self._userSet = True
        return ans

    def value(self, accessValue=True):
        if accessValue:
            self._userAccessed = True
        return dict((name, config.value(accessValue))
                    for name, config in six.iteritems(self._data))

    def set_value(self, value, skip_implicit=False):
        if value is None:
            return self
        if (type(value) is not dict) and \
           (not isinstance(value, ConfigDict)):
            raise ValueError("Expected dict value for %s.set_value, found %s" %
                             (self.name(True), type(value).__name__))
        if not value:
            return self
        _implicit = []
        _decl_map = {}
        for key in value:
            _key = str(key)
            if _key in self._data:
                # str(key) may not be key... store the mapping so that
                # when we later iterate over the _decl_order, we can map
                # the local keys back to the incoming value keys.
                _decl_map[_key] = key
            else:
                _key = _key.replace('_', ' ')
                if _key in self._data:
                    _decl_map[str(_key)] = key
                else:
                    if skip_implicit:
                        pass
                    elif self._implicit_declaration:
                        _implicit.append(key)
                    else:
                        raise ValueError(
                            "key '%s' not defined for Config Block '%s' and "
                            "implicit (undefined) keys are not allowed" %
                            (key, self.name(True)))

        # If the set_value fails part-way through the new values, we
        # want to restore a deterministic state.  That is, either
        # set_value succeeds completely, or else nothing happens.
        _old_data = self.value(False)
        try:
            # We want to set the values in declaration order (so that
            # things are deterministic and in case a validation depends
            # on the order)
            for key in self._decl_order:
                if key in _decl_map:
                    #print "Setting", key, " = ", value
                    self[key] = value[_decl_map[key]]
            # implicit data is declared at the end (in sorted order)
            for key in sorted(_implicit):
                self.add(key, value[key])
        except:
            self.reset()
            self.set_value(_old_data)
            raise
        self._userSet = True
        return self

    def reset(self):
        # Reset the values in the order they were declared.  This
        # allows reset functions to have a deterministic ordering.
        def _keep(self, key):
            keep = key in self._declared
            if keep:
                self._data[key].reset()
            else:
                del self._data[key]
            return keep
        # this is an in-place slice of a list...
        self._decl_order[:] = [x for x in self._decl_order if _keep(self, x)]
        self._userAccessed = False
        self._userSet = False

    def _data_collector(self, level, prefix, visibility=None, docMode=False):
        if visibility is not None and visibility < self._visibility:
            return
        if prefix:
            yield (level, prefix, None, self)
            if level is not None:
                level += 1
        for key in self._decl_order:
            for v in self._data[key]._data_collector(level, key + ': ',
                                                     visibility, docMode):
                yield v

# Backwards compatibility: ConfigDick was originally named ConfigBlock.
ConfigBlock = ConfigDict

# In Python3, the items(), etc methods of dict-like things return
# generator-like objects.
if six.PY3:
    ConfigDict.keys = ConfigDict.iterkeys
    ConfigDict.values = ConfigDict.itervalues
    ConfigDict.items = ConfigDict.iteritems


class In(object):
    """A ConfigValue domain validator that checks values against a set

    Instances of In map incoming values to the desired type (if domain
    is specified) and check that the resulting value is in the specified
    set.

    Examples
    --------
    >>> c = ConfigValue(domain=In(['foo', 'bar', '0'], domain=str))
    >>> c.set_value('foo')
    >>> c.display
    foo
    >>> c.set_value(3)
    ValueError: invalid value for configuration '':
            Failed casting 3
            to <class 'pyutilib.misc.config.In'>
            Error: value 3 not in domain ['foo', 'bar']
    >>> c.display
    foo
    >>> c.set_value(0)
    >>> c.display
    '0'

    """

    def __init__(self, allowable, domain=None):
        self._allowable = allowable
        self._domain = domain

    def __call__(self, value):
        if self._domain is not None:
            v = self._domain(value)
        else:
            v = value
        if v in self._allowable:
            return v
        raise ValueError("value %s not in domain %s" % (value, self._allowable))



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________


class Factory(object):
    """
    A class that is used to define a factory for objects.

    Factory objects may be cached for future use.
    """

    def __init__(self, description=None):
        self._description = description
        self._cls = {}
        self._doc = {}
        self._cached = False
        self._cache = {}

    def __call__(self, name, **kwds):
        if 'exception' in kwds:
            exception = kwds['exception']
            del kwds['exception']
        else:
            exception = False
        name = str(name)
        if not name in self._cls:
            if not exception:
                return None
            if self._description is None:
                raise ValueError("Unknown factory object type: '%s'" % name)
            raise ValueError("Unknown %s: '%s'" % (self._description, name))
        if self._cached:
            if name not in self._cache:
                self._cache[name] = self._cls[name](**kwds)
            return self._cache[name]
        return self._cls[name](**kwds)

    def __iter__(self):
        for name in self._cls:
            yield name

    def __contains__(self, name):
        return str(name) in self._cls

    def get_class(self, name):
        return self._cls[name]

    def doc(self, name):
        return self._doc[name]

    def unregister(self, name):
        name = str(name)
        if name in self._cls:
            del self._cls[name]
            del self._doc[name]
    
    def register(self, name, doc=None):
        def fn(cls):
            self._cls[name] = cls
            self._doc[name] = doc
            return cls
        return fn

    #
    # The following methods might be used by developers working
    # with factories defined using PyUtilib component plugins.  These methods
    # are deprecated.
    #
    def services(self):
        raise RuntimeError("ERROR: Factory.services() is not defined.  Use Factory.__iter__() instead.")

    def activate(self):
        raise RuntimeError("ERROR: Factory.activate() is not defined.")

    def deactivate(self):
        raise RuntimeError("ERROR: Factory.deactivate() is not defined.")


class CachedFactory(Factory):
    """
    A class that is used to define a factory for objects.

    Factory objects are cached for future use.
    """

    def __init__(self, description=None):
        Factory.__init__(self, description)
        self._cached = True




#
# Unit Tests for misc/archivereader
#
#

import os
import fnmatch
import posixpath

import pyutilib.th as unittest

from pyutilib.misc.archivereader import \
    ArchiveReaderFactory, DirArchiveReader, FileArchiveReader, \
    TarArchiveReader, ZipArchiveReader, BZ2FileArchiveReader, \
    GzipFileArchiveReader, tarfile_available, zipfile_available, \
    gzip_available, bz2_available

currdir = os.path.dirname(os.path.abspath(__file__))
testdatadir = os.path.join(currdir, 'archivereader')


class TestArchiveReaderFactory(unittest.TestCase):

    def test_ArchiveReaderFactory_dir(self):
        archive = ArchiveReaderFactory(
            os.path.join(testdatadir, 'archive_flat'))
        self.assertTrue(isinstance(archive, DirArchiveReader))

    @unittest.skipUnless(tarfile_available, "tarfile support is disabled")
    def test_ArchiveReaderFactory_tar(self):
        archive = ArchiveReaderFactory(
            os.path.join(testdatadir, 'archive_flat.tar'))
        self.assertTrue(isinstance(archive, TarArchiveReader))

    @unittest.skipUnless(tarfile_available, "tarfile support is disabled")
    def test_ArchiveReaderFactory_targz(self):
        archive = ArchiveReaderFactory(
            os.path.join(testdatadir, 'archive_flat.tar.gz'))
        self.assertTrue(isinstance(archive, TarArchiveReader))

    @unittest.skipUnless(tarfile_available, "tarfile support is disabled")
    def test_ArchiveReaderFactory_tgz(self):
        archive = ArchiveReaderFactory(
            os.path.join(testdatadir, 'archive_flat.tgz'))
        self.assertTrue(isinstance(archive, TarArchiveReader))

    @unittest.skipUnless(zipfile_available, "zipfile support is disabled")
    def test_ArchiveReaderFactory_zip(self):
        archive = ArchiveReaderFactory(
            os.path.join(testdatadir, 'archive_flat.zip'))
        self.assertTrue(isinstance(archive, ZipArchiveReader))

    def test_ArchiveReaderFactory_file(self):
        archive = ArchiveReaderFactory(os.path.join(testdatadir, 'fileC.txt'))
        self.assertTrue(isinstance(archive, FileArchiveReader))

    @unittest.skipUnless(gzip_available, "gzip support is disabled")
    def test_ArchiveReaderFactory_gzip(self):
        archive = ArchiveReaderFactory(
            os.path.join(testdatadir, 'fileC.txt.gz'))
        self.assertTrue(isinstance(archive, GzipFileArchiveReader))

    @unittest.skipUnless(bz2_available, "bz2 support is disabled")
    def test_ArchiveReaderFactory_bzip2(self):
        archive = ArchiveReaderFactory(
            os.path.join(testdatadir, 'fileC.txt.bz2'))
        self.assertTrue(isinstance(archive, BZ2FileArchiveReader))

    def test_ArchiveReaderFactory_nonexist(self):
        try:
            archive = ArchiveReaderFactory(
                os.path.join(testdatadir, '_does_not_exist_'))
        except IOError:
            pass
        else:
            self.fail(
                "ArchiveReaderFactory should raise exception with nonexistent archive")


class _TestArchiveReaderBaseNested(object):

    archive_class = None
    archive_name = None
    archive_class_kwds = {}

    def setUp(self):
        self._a = None

    def tearDown(self):
        if self._a is not None:
            self._a.close()

    def test_default(self):
        a = self._a = self.archive_class(
            os.path.join(testdatadir, self.archive_name),
            **self.archive_class_kwds)
        tmpdir = a.extract('directory', recursive=True)
        self.assertEqual(os.path.exists(tmpdir), True)
        self.assertEqual(
            os.path.exists(os.path.join(tmpdir, 'fileA.txt')), True)
        self.assertEqual(
            os.path.exists(os.path.join(tmpdir, 'fileB.txt')), True)
        a.clear_extractions()
        self.assertEqual(os.path.exists(tmpdir), False)
        self.assertEqual(
            os.path.exists(os.path.join(tmpdir, 'fileA.txt')), False)
        self.assertEqual(
            os.path.exists(os.path.join(tmpdir, 'fileB.txt')), False)

        tmpdir = a.extract('directory', recursive=True)
        self.assertEqual(os.path.exists(tmpdir), True)
        self.assertEqual(
            os.path.exists(os.path.join(tmpdir, 'fileA.txt')), True)
        self.assertEqual(
            os.path.exists(os.path.join(tmpdir, 'fileB.txt')), True)
        a.clear_extractions()
        self.assertEqual(os.path.exists(tmpdir), False)
        self.assertEqual(
            os.path.exists(os.path.join(tmpdir, 'fileA.txt')), False)
        self.assertEqual(
            os.path.exists(os.path.join(tmpdir, 'fileB.txt')), False)

        tmpnames = a.extractall()
        for name in tmpnames:
            self.assertEqual(os.path.exists(name), True)
        names = [a.normalize_name(name).replace(
            a.normalize_name(a._workdir) + '/', '') for name in tmpnames]
        self.assertEqual(
            sorted(names),
            sorted(['directory', posixpath.join('directory', 'fileA.txt'),
                    posixpath.join('directory', 'fileB.txt')]))
        a.clear_extractions()
        for name in tmpnames:
            self.assertEqual(os.path.exists(name), False)
        self.assertEqual(
            sorted(a.getnames()),
            sorted(['directory', posixpath.join('directory', 'fileA.txt'),
                    posixpath.join('directory', 'fileB.txt')]))
        f = a.open(posixpath.join('directory', 'fileA.txt'))
        try:
            self.assertEqual(f.read().strip().decode(), 'this is fileA')
        finally:
            f.close()
        f = a.open(posixpath.join('directory', 'fileB.txt'))
        try:
            self.assertEqual(f.read().strip().decode(), 'this is fileB')
        finally:
            f.close()
        a.close()

    def test_maxdepth0(self):
        a = self._a = self.archive_class(
            os.path.join(testdatadir, self.archive_name),
            maxdepth=0,
            **self.archive_class_kwds)
        tmpdir = a.extract('directory', recursive=True)
        self.assertEqual(os.path.exists(tmpdir), True)
        self.assertEqual(
            os.path.exists(os.path.join(tmpdir, 'fileA.txt')), False)
        self.assertEqual(
            os.path.exists(os.path.join(tmpdir, 'fileB.txt')), False)
        a.clear_extractions()
        self.assertEqual(os.path.exists(tmpdir), False)
        self.assertEqual(
            os.path.exists(os.path.join(tmpdir, 'fileA.txt')), False)
        self.assertEqual(
            os.path.exists(os.path.join(tmpdir, 'fileB.txt')), False)

        tmpnames = a.extractall()
        for name in tmpnames:
            self.assertEqual(os.path.exists(name), True)
        names = [a.normalize_name(name).replace(
            a.normalize_name(a._workdir) + '/', '') for name in tmpnames]
        self.assertEqual(sorted(names), sorted(['directory']))
        a.clear_extractions()
        for name in tmpnames:
            self.assertEqual(os.path.exists(name), False)

        self.assertEqual(sorted(a.getnames()), sorted(['directory']))
        a.close()

    def test_maxdepth1(self):
        a = self._a = self.archive_class(
            os.path.join(testdatadir, self.archive_name),
            maxdepth=1,
            **self.archive_class_kwds)

        tmpdir = a.extract('directory', recursive=True)
        self.assertEqual(os.path.exists(tmpdir), True)
        self.assertEqual(
            os.path.exists(os.path.join(tmpdir, 'fileA.txt')), True)
        self.assertEqual(
            os.path.exists(os.path.join(tmpdir, 'fileB.txt')), True)
        a.clear_extractions()
        self.assertEqual(os.path.exists(tmpdir), False)
        self.assertEqual(
            os.path.exists(os.path.join(tmpdir, 'fileA.txt')), False)
        self.assertEqual(
            os.path.exists(os.path.join(tmpdir, 'fileB.txt')), False)

        tmpnames = a.extractall()
        for name in tmpnames:
            self.assertEqual(os.path.exists(name), True)
        names = [a.normalize_name(name).replace(
            a.normalize_name(a._workdir) + '/', '') for name in tmpnames]
        self.assertEqual(
            sorted(names),
            sorted(['directory', posixpath.join('directory', 'fileA.txt'),
                    posixpath.join('directory', 'fileB.txt')]))
        a.clear_extractions()
        for name in tmpnames:
            self.assertEqual(os.path.exists(name), False)

        self.assertEqual(
            sorted(a.getnames()),
            sorted(['directory', posixpath.join('directory', 'fileA.txt'),
                    posixpath.join('directory', 'fileB.txt')]))
        f = a.open(posixpath.join('directory', 'fileA.txt'))
        try:
            self.assertEqual(f.read().strip().decode(), 'this is fileA')
        finally:
            f.close()
        f = a.open(posixpath.join('directory', 'fileB.txt'))
        try:
            self.assertEqual(f.read().strip().decode(), 'this is fileB')
        finally:
            f.close()
        a.close()

    def test_subdir_maxdepth0(self):
        a = self._a = self.archive_class(
            os.path.join(testdatadir, self.archive_name),
            subdir='directory',
            maxdepth=0,
            **self.archive_class_kwds)
        tmpnames = a.extractall()
        for name in tmpnames:
            self.assertEqual(os.path.exists(name), True)
        names = [a.normalize_name(name).replace(
            posixpath.join(a.normalize_name(a._workdir), a._subdir), '')
                 for name in tmpnames]
        self.assertEqual(sorted(names), sorted(['fileA.txt', 'fileB.txt']))
        a.clear_extractions()
        for name in tmpnames:
            self.assertEqual(os.path.exists(name), False)
        a.close()
        a = self._a = self.archive_class(
            os.path.join(testdatadir, self.archive_name),
            subdir='directory',
            maxdepth=0,
            **self.archive_class_kwds)
        self.assertEqual(
            sorted(a.getnames()), sorted(['fileA.txt', 'fileB.txt']))
        f = a.open('fileA.txt')
        try:
            self.assertEqual(f.read().strip().decode(), 'this is fileA')
        finally:
            f.close()
        f = a.open('fileB.txt')
        try:
            self.assertEqual(f.read().strip().decode(), 'this is fileB')
        finally:
            f.close()
        a.close()

    def test_subdir_maxdepth1(self):
        a = self._a = self.archive_class(
            os.path.join(testdatadir, self.archive_name),
            subdir='directory',
            maxdepth=1,
            **self.archive_class_kwds)
        tmpnames = a.extractall()
        for name in tmpnames:
            self.assertEqual(os.path.exists(name), True)
        names = [a.normalize_name(name).replace(
            posixpath.join(a.normalize_name(a._workdir), a._subdir), '')
                 for name in tmpnames]
        self.assertEqual(sorted(names), sorted(['fileA.txt', 'fileB.txt']))
        a.clear_extractions()
        for name in tmpnames:
            self.assertEqual(os.path.exists(name), False)

        self.assertEqual(
            sorted(a.getnames()), sorted(['fileA.txt', 'fileB.txt']))
        f = a.open('fileA.txt')
        try:
            self.assertEqual(f.read().strip().decode(), 'this is fileA')
        finally:
            f.close()
        f = a.open('fileB.txt')
        try:
            self.assertEqual(f.read().strip().decode(), 'this is fileB')
        finally:
            f.close()
        a.close()


def dir_svn_junk_filter(x):
    return (fnmatch.fnmatch(x,"*.junk") or \
            fnmatch.fnmatch(x,"*.svn*"))


class TestDirArchiveReader(_TestArchiveReaderBaseNested, unittest.TestCase):

    archive_class = DirArchiveReader
    archive_name = 'archive_directory'
    archive_class_kwds = {'filter': dir_svn_junk_filter}


class TestDirArchiveReaderWin(_TestArchiveReaderBaseNested, unittest.TestCase):

    archive_class = DirArchiveReader
    archive_name = 'win_archive_directory'
    archive_class_kwds = {'filter': dir_svn_junk_filter}


@unittest.skipUnless(tarfile_available, "tarfile support is disabled")
class TestTarArchiveReader1(_TestArchiveReaderBaseNested, unittest.TestCase):

    archive_class = TarArchiveReader
    archive_name = 'archive_directory.tgz'


@unittest.skipUnless(tarfile_available, "tarfile support is disabled")
class TestTarArchiveReader2(_TestArchiveReaderBaseNested, unittest.TestCase):

    archive_class = TarArchiveReader
    archive_name = 'archive_directory.tar.gz'


@unittest.skipUnless(tarfile_available, "tarfile support is disabled")
class TestTarArchiveReader3(_TestArchiveReaderBaseNested, unittest.TestCase):

    archive_class = TarArchiveReader
    archive_name = 'archive_directory.tar'


@unittest.skipUnless(zipfile_available, "zipfile support is disabled")
class TestZipArchiveReader(_TestArchiveReaderBaseNested, unittest.TestCase):

    archive_class = ZipArchiveReader
    archive_name = 'archive_directory.zip'


@unittest.skipUnless(zipfile_available, "zipfile support is disabled")
class TestZipArchiveReaderWin(_TestArchiveReaderBaseNested, unittest.TestCase):

    archive_class = ZipArchiveReader
    archive_name = 'win_archive_directory.zip'


class _TestFileArchiveReaderBase(object):

    archive_class = None
    archive_name = None

    def setUp(self):
        self._a = None

    def tearDown(self):
        if self._a is not None:
            self._a.close()

    def test1(self):
        a = self._a = self.archive_class(
            os.path.join(testdatadir, self.archive_name))
        f = a.open('fileC.txt')
        try:
            self.assertEqual(f.read().strip().decode(), 'this is fileC')
        finally:
            f.close()
        f = a.open()
        try:
            self.assertEqual(f.read().strip().decode(), 'this is fileC')
        finally:
            f.close()
        dst = a.extract()
        self.assertEqual(os.path.exists(dst), True)
        with open(dst, 'rb') as f:
            self.assertEqual(f.read().strip().decode(), 'this is fileC')
        a.clear_extractions()
        self.assertEqual(os.path.exists(dst), False)
        dst = a.extract('fileC.txt')
        self.assertEqual(os.path.exists(dst), True)
        with open(dst, 'rb') as f:
            self.assertEqual(f.read().strip().decode(), 'this is fileC')
        a.clear_extractions()
        self.assertEqual(os.path.exists(dst), False)
        a.close()


@unittest.skipUnless(gzip_available, "gzip support is disabled")
class TestGzipFileArchiveReader(_TestFileArchiveReaderBase, unittest.TestCase):

    archive_class = GzipFileArchiveReader
    archive_name = 'fileC.txt.gz'


@unittest.skipUnless(bz2_available, "bz2 support is disabled")
class TestBZ2FileArchiveReader(_TestFileArchiveReaderBase, unittest.TestCase):

    archive_class = BZ2FileArchiveReader
    archive_name = 'fileC.txt.bz2'


class TestFileArchiveReader(_TestFileArchiveReaderBase, unittest.TestCase):

    archive_class = FileArchiveReader
    archive_name = 'fileC.txt'


if __name__ == "__main__":
    unittest.main()



from pyutilib.misc.timing import HierarchicalTimer
import pyutilib.th as unittest


class TestHierarchicalTimer(unittest.TestCase):
    def test_hierarchical_timer(self):
        timer = HierarchicalTimer()
        timer.start('all')
        for i in range(10):
            timer.start('a')
            for i in range(5):
                timer.start('aa')
                timer.stop('aa')
            timer.start('ab')
            timer.stop('ab')
            timer.stop('a')
            timer.start('b')
            timer.stop('b')
        timer.start('a')
        with self.assertRaisesRegex(ValueError, 'all is not the currently active timer.  The only timer that can currently be stopped is all.a'):
            timer.stop('all')
        timer.stop('a')
        timer.stop('all')

        a_percent = timer.get_relative_percent_time('all.a')
        aa_percent = timer.get_relative_percent_time('all.a.aa')
        aa_total_percent = timer.get_total_percent_time('all.a.aa')
        self.assertAlmostEqual(aa_total_percent, a_percent/100 * aa_percent/100 * 100)
        self.assertAlmostEqual(timer.get_num_calls('all.a'), 11)
        self.assertAlmostEqual(timer.get_num_calls('all.a.ab'), 10)
        self.assertAlmostEqual(timer.get_num_calls('all.a.aa'), 50)
        timer.get_total_time('all.b')
        print(timer)



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
"""A parser used by the Configuration class."""

__all__ = ['Configuration_ConfigParser']

import os.path
try:
    from ordereddict import OrderedDict
except:
    OrderedDict = dict
from pyutilib.component.core import implements
from pyutilib.component.config.configuration import ConfigurationError, IConfiguration
from pyutilib.component.config.managed_plugin import ManagedSingletonPlugin
from pyutilib.component.config.options import declare_option

def _configParser(**kwds):
    try:
        import ConfigParser
    except ImportError:
        import configparser as ConfigParser
    #
    # Force the config file option manager to be case sensitive
    #
    ConfigParser.RawConfigParser.optionxform = str
    return ConfigParser.ConfigParser(**kwds)

class Configuration_ConfigParser(ManagedSingletonPlugin):
    """A configuration parser that uses the ConfigParser package."""

    implements(IConfiguration)

    def __init__(self, **kwds):
        kwds['name'] = 'Configuration_ConfigParser'
        ManagedSingletonPlugin.__init__(self, **kwds)

    def load(self, filename):
        """Returns a list of tuples: [ (section,option,value) ]"""
        parser = _configParser()
        if not os.path.exists(filename):
            raise ConfigurationError("File " + filename + " does not exist!")
        parser.read(filename)
        #
        # Collect data
        #
        data = []
        for section in parser.sections():
            for (option, value) in parser.items(section):
                data.append((section, option, value))
        return data

    def save(self, filename, config, header=None):
        """Save configuration information to the specified file."""
        parser = _configParser()
        for (section, option, value) in config:
            if not parser.has_section(section):
                parser.add_section(section)
            parser.set(section, option, str(value))
        OUTPUT = open(filename, "w")
        if not header is None:
            for line in header.split("\n"):
                OUTPUT.write("; " + line + '\n')
        parser.write(OUTPUT)
        OUTPUT.close()



#
# Unit Tests for xml_utils
#
#

import os
from os.path import abspath, dirname
pkgdir = dirname(abspath(__file__))
currdir = dirname(abspath(__file__)) + os.sep
import pyutilib.th as unittest
import pyutilib.misc

from xml.dom import minidom


class XMLDebug(unittest.TestCase):

    def setUp(self):
        self.doc = minidom.parse(pkgdir + "/test.xml")
        self.node = self.doc.documentElement

    def tearDown(self):
        self.doc = None
        self.node = None

    def test_get_text(self):
        # Verify that we can get XML text
        str = pyutilib.misc.get_xml_text(self.node)
        target = "a b c\n  \n  d e f"
        self.assertEqual(target, str)

    def test_escape(self):
        source = "&'<>\""
        str = pyutilib.misc.escape(source)
        target = "&amp;'&lt;&gt;\""
        self.assertEqual(target, str)


class Test(unittest.TestCase):

    def test1(self):
        # Verify that data 1 is not equal to data 2
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_xml_files,
            currdir + 'xmldata1.xml',
            currdir + 'xmldata2.xml',
            exact=True)

    def test2(self):
        # Verify that data 1 is a subset of data 2
        pyutilib.misc.compare_xml_files(
            currdir + 'xmldata1.xml', currdir + 'xmldata2.xml', exact=False)

    def test3(self):
        # Verify that data 2 is not a subset of data 1
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_xml_files,
            currdir + 'xmldata2.xml',
            currdir + 'xmldata1.xml',
            exact=True)

    def test4(self):
        # Verify check to confirm that YAML data has a different type
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_xml_files,
            currdir + 'xmldata1.xml',
            currdir + 'xmldata3.xml',
            exact=True)

    def test5(self):
        # Verify that data 1 is not equal to data 1
        pyutilib.misc.compare_xml_files(
            currdir + 'xmldata1.xml', currdir + 'xmldata1.xml', exact=True)

    def test6(self):
        # Verify that data 1 is a subset of data 4
        pyutilib.misc.compare_xml_files(
            currdir + 'xmldata1.xml',
            currdir + 'xmldata4.txt',
            output_begin='BEGIN',
            output_end='END',
            exact=False)

    def test7(self):
        # Verify that data 6 is a subset of data 5
        pyutilib.misc.compare_xml_files(
            currdir + 'xmldata6.xml', currdir + 'xmldata5.xml', exact=False)

    def test8(self):
        # Verify that data 5 is not a subset of data 6
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_xml_files,
            currdir + 'xmldata5.xml',
            currdir + 'xmldata6.xml',
            exact=False)

    def test9(self):
        # Verify that data 7 is not a subset of data 5
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_xml_files,
            currdir + 'xmldata7.xml',
            currdir + 'xmldata5.xml',
            exact=False)

    def test10(self):
        # Verify that data 6 is not equal to data 5
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_xml_files,
            currdir + 'xmldata6.xml',
            currdir + 'xmldata5.xml',
            exact=True)

    def test11(self):
        # Verify that data 8 is not equal to data 5
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_xml_files,
            currdir + 'xmldata8.xml',
            currdir + 'xmldata5.xml',
            exact=True)

    def test12(self):
        # Verify that data 9 is not a subset of data 1
        self.assertRaises(
            IOError,
            pyutilib.misc.compare_xml_files,
            currdir + 'xmldata9.xml',
            currdir + 'xmldata1.xml',
            exact=True)

    def test13(self):
        # Verify that data 10 is not a subset of data 1
        self.assertRaises(
            ValueError,
            pyutilib.misc.compare_xml_files,
            currdir + 'xmldata10.xml',
            currdir + 'xmldata1.xml',
            exact=True)

    def test14(self):
        # Verify that data 10 is a subset of data 1 with a loose tolerance
        pyutilib.misc.compare_xml_files(
            currdir + 'xmldata10.xml',
            currdir + 'xmldata1.xml',
            tolerance=1.0,
            exact=True)


if __name__ == "__main__":
    unittest.main()
